import re
from pathlib import Path

import pdoc
from changelog_gen.context import Context


def generate_docs(ctx: Context, _new: str) -> list[str]:
    output_dir = Path("./docs")
    modules = ctx.config.custom["pdoc"]["modules"]

    context = pdoc.Context()

    modules = [
        pdoc.Module(mod, context=context, skip_errors=ctx.config.custom["pdoc"].get("skip_errors", False))
        for mod in modules
    ]
    pdoc.link_inheritance(context)

    def recursive_mds(mod: pdoc.Module) -> pdoc.Module:
        yield mod
        for submod in mod.submodules():
            yield from recursive_mds(submod)

    paths = []

    for mod in modules:
        for module in recursive_mds(mod):
            path = re.sub(r"\.html$", ".md", module.url())
            out = output_dir / path
            out.parent.mkdir(exist_ok=True, parents=True)
            with out.open("w") as f:
                f.write(module.text(show_inherited_members=False))
            paths.append(str(out))

    return paths



import importlib
import inspect

from fastapi import FastAPI

from neos_common import schema, util


def get_error_codes(module_names: str | list[str]) -> schema.ErrorCodes:
    """Get all error codes and messages from the error module."""
    errors = []

    if isinstance(module_names, str):
        module_names = [module_names]

    for module_name in module_names:
        module = importlib.import_module(module_name)

        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj):
                if hasattr(obj, "title"):
                    o = obj("details")
                    type_ = o.type
                    title = o.title

                    errors.append(
                        schema.ErrorCode(
                            class_name=name,
                            type=type_,
                            title=title,
                        ),
                    )

                elif hasattr(obj, "reason_"):
                    o = obj("debug")
                    type_ = o.reason
                    title = o.message

                    errors.append(
                        schema.ErrorCode(
                            class_name=name,
                            type=type_,
                            title=title,
                        ),
                    )

    return schema.ErrorCodes(errors=errors)


def get_permissions(app: FastAPI, ignore_routes: list[str] | None = None) -> schema.FormattedRoutes:
    ignore_routes = ignore_routes or []
    routes = util.get_routes(app, ignore_routes)

    return schema.FormattedRoutes(routes=routes)



import abc
import datetime
import enum
import json
import logging
import random
import re
import string
import time
import typing
import uuid
from pathlib import Path
from typing import Annotated

import pydantic

EMIT_LOG_FILENAME = "emit.log"
MAX_EMIT_LOG_LINE_LENGTH = 1000
ROOT_DIR = "/tmp/core-events"  # noqa: S108

logger = logging.getLogger(__name__)

UUID4 = Annotated[pydantic.UUID4, pydantic.PlainSerializer(lambda x: str(x))]


class EventPacket(pydantic.BaseModel):
    source: str
    timestamp: int  # timestamp in ms
    span_id: UUID4
    version: str
    message: str | dict[str, typing.Any]
    message_type: str
    session_id: UUID4 | None = None


class EventPackets(pydantic.BaseModel):
    packets: list[EventPacket]


class EventVersion:
    VERSION_PATTERN = re.compile("^v(\\d+).(\\d+).(\\d+)$")

    def __init__(self, major: int, minor: int, patch: int) -> None:
        """Define version of the event.

        Args:
            major (int): version of the Gateway API
            minor (int): breaking changes
            patch (int): non-breaking changes
        """
        self.major = major
        self.minor = minor
        self.patch = patch

    def __eq__(self, other: object) -> bool:
        """Compare version for equality."""
        if not isinstance(other, EventVersion):
            return NotImplemented
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __le__(self, other: "EventVersion") -> bool:
        """Compare version."""
        return (self.major, self.minor, self.patch) <= (other.major, other.minor, other.patch)

    def __str__(self) -> str:
        """Convert to string."""
        return f"v{self.major}.{self.minor}.{self.patch}"

    @staticmethod
    def from_string(value: str) -> "EventVersion":
        m = EventVersion.VERSION_PATTERN.match(value)
        if not m:
            msg: str = f"Could not parse event version from {value}"
            raise ValueError(msg)
        return EventVersion(int(m.group(1)), int(m.group(2)), int(m.group(3)))


class EventType(enum.Enum):
    secret = "secret"  # noqa: S105
    tag = "tag"
    output = "output"
    data_product = "data_product"
    data_source = "data_source"
    data_system = "data_system"
    data_unit = "data_unit"
    spark = "spark"
    builder = "builder"
    journal_note = "journal_note"


class EventScope(enum.Enum):
    secret = EventType.secret.value
    secret_keys = f"{EventType.secret.value}.keys"

    tag = EventType.tag.value

    output = EventType.output.value
    output_info = f"{EventType.output.value}.info"
    output_journal = f"{EventType.output.value}.journal"

    data_product = EventType.data_product.value
    data_product_data = f"{EventType.data_product.value}.data"
    data_product_metadata = f"{EventType.data_product.value}.metadata"
    data_product_output = f"{EventType.data_product.value}.output"
    data_product_data_product = f"{EventType.data_product.value}.data_product"
    data_product_info = f"{EventType.data_product.value}.info"
    data_product_journal = f"{EventType.data_product.value}.journal"
    data_product_schema = f"{EventType.data_product.value}.schema"
    data_product_spark = f"{EventType.data_product.value}.spark"
    data_product_spark_state = f"{EventType.data_product.value}.spark_state"
    data_product_spark_builder = f"{EventType.data_product.value}.spark_builder"
    data_product_expectations = f"{EventType.data_product.value}.expectations"
    data_product_data_quality = f"{EventType.data_product.value}.data_quality"
    data_product_profiling = f"{EventType.data_product.value}.profiling"
    data_product_validation = f"{EventType.data_product.value}.validation"
    data_product_classification = f"{EventType.data_product.value}.classification"
    data_product_classification_configuration = f"{EventType.data_product.value}.classification.configuration"
    data_product_classification_result = f"{EventType.data_product.value}.classification.result"
    data_product_health = f"{EventType.data_product.value}.health"

    data_source = EventType.data_source.value
    data_source_info = f"{EventType.data_source.value}.info"
    data_source_journal = f"{EventType.data_source.value}.journal"
    data_source_connection = f"{EventType.data_source.value}.connection"
    data_source_health = f"{EventType.data_source.value}.health"
    data_source_secret = f"{EventType.data_source.value}.secret"
    data_source_state = f"{EventType.data_source.value}.state"
    data_source_data_unit = f"{EventType.data_source.value}.data_unit"

    data_system = EventType.data_system.value
    data_system_info = f"{EventType.data_system.value}.info"
    data_system_journal = f"{EventType.data_system.value}.journal"
    data_system_data_source = f"{EventType.data_system.value}.data_source"

    data_unit = EventType.data_unit.value
    data_unit_metadata = f"{EventType.data_unit.value}.metadata"
    data_unit_schema = f"{EventType.data_unit.value}.schema"
    data_unit_state = f"{EventType.data_unit.value}.state"
    data_unit_info = f"{EventType.data_unit.value}.info"
    data_unit_journal = f"{EventType.data_unit.value}.journal"
    data_unit_health = f"{EventType.data_unit.value}.health"
    data_unit_profiling = f"{EventType.data_unit.value}.profiling"
    data_unit_config = f"{EventType.data_unit.value}.config"
    data_unit_data_product = f"{EventType.data_unit.value}.data_product"

    spark = EventType.spark.value
    spark_status = f"{EventType.spark.value}.status"

    builder = EventType.builder.value

    journal_note = EventType.journal_note.value


class EventAction(enum.Enum):
    secret_add = f"{EventScope.secret.value}:add"
    secret_update = f"{EventScope.secret.value}:update"
    secret_delete = f"{EventScope.secret.value}:delete"
    secret_keys_delete = f"{EventScope.secret_keys.value}:delete"

    tag_add = f"{EventScope.tag.value}:add"
    tag_delete = f"{EventScope.tag.value}:delete"

    output_create = f"{EventScope.output.value}:create"
    output_update = f"{EventScope.output.value}:update"
    output_delete = f"{EventScope.output.value}:delete"
    output_info_update = f"{EventScope.output_info.value}:update"
    output_journal_add = f"{EventScope.output_journal.value}:add"

    data_product_create = f"{EventScope.data_product.value}:create"
    data_product_update = f"{EventScope.data_product.value}:update"
    data_product_delete = f"{EventScope.data_product.value}:delete"
    data_product_data_delete = f"{EventScope.data_product_data.value}:delete"
    data_product_metadata_update = f"{EventScope.data_product_metadata.value}:update"
    data_product_metadata_remove = f"{EventScope.data_product_metadata.value}:remove"
    data_product_schema_update = f"{EventScope.data_product_schema.value}:update"
    data_product_output_link = f"{EventScope.data_product_output.value}:link"
    data_product_output_unlink = f"{EventScope.data_product_output.value}:unlink"
    data_product_data_product_link = f"{EventScope.data_product_data_product.value}:link"
    data_product_data_product_unlink = f"{EventScope.data_product_data_product.value}:unlink"
    data_product_info_update = f"{EventScope.data_product_info.value}:update"
    data_product_journal_add = f"{EventScope.data_product_journal.value}:add"
    data_product_publish = f"{EventScope.data_product.value}:publish"
    data_product_unpublish = f"{EventScope.data_product.value}:unpublish"
    data_product_spark_update = f"{EventScope.data_product_spark.value}:update"
    data_product_spark_builder_update = f"{EventScope.data_product_spark_builder.value}:update"
    data_product_spark_state_update = f"{EventScope.data_product_spark_state.value}:update"
    data_product_expectations_update = f"{EventScope.data_product_expectations.value}:update"
    data_product_profiling_update = f"{EventScope.data_product_profiling.value}:update"
    data_product_validation_update = f"{EventScope.data_product_validation.value}:update"
    data_product_classification_update = f"{EventScope.data_product_classification.value}:update"
    data_product_classification_configuration_update = (
        f"{EventScope.data_product_classification_configuration.value}:update"
    )
    data_product_classification_result_update = f"{EventScope.data_product_classification_result.value}:update"
    data_product_health_check = f"{EventScope.data_product_health.value}:check"

    data_source_create = f"{EventScope.data_source.value}:create"
    data_source_update = f"{EventScope.data_source.value}:update"
    data_source_delete = f"{EventScope.data_source.value}:delete"
    data_source_info_update = f"{EventScope.data_source_info.value}:update"
    data_source_journal_add = f"{EventScope.data_source_journal.value}:add"
    data_source_connection_update = f"{EventScope.data_source_connection.value}:update"
    data_source_secret_add = f"{EventScope.data_source_secret.value}:add"
    data_source_data_unit_link = f"{EventScope.data_source_data_unit.value}:link"
    data_source_data_unit_unlink = f"{EventScope.data_source_data_unit.value}:unlink"
    data_source_health_check = f"{EventScope.data_source_health.value}:check"
    data_source_state_update = f"{EventScope.data_source_state.value}:update"

    data_system_create = f"{EventScope.data_system.value}:create"
    data_system_update = f"{EventScope.data_system.value}:update"
    data_system_delete = f"{EventScope.data_system.value}:delete"
    data_system_info_update = f"{EventScope.data_system_info.value}:update"
    data_system_info_add = f"{EventScope.data_system_info.value}:add"
    data_system_journal_add = f"{EventScope.data_system_journal.value}:add"
    data_system_data_source_link = f"{EventScope.data_system_data_source.value}:link"
    data_system_data_source_unlink = f"{EventScope.data_system_data_source.value}:unlink"

    data_unit_create = f"{EventScope.data_unit.value}:create"
    data_unit_update = f"{EventScope.data_unit.value}:update"
    data_unit_metadata_update = f"{EventScope.data_unit_metadata.value}:update"
    data_unit_metadata_delete = f"{EventScope.data_unit_metadata.value}:delete"
    data_unit_schema_update = f"{EventScope.data_unit_schema.value}:update"
    data_unit_state_update = f"{EventScope.data_unit_state.value}:update"
    data_unit_delete = f"{EventScope.data_unit.value}:delete"
    data_unit_info_update = f"{EventScope.data_unit_info.value}:update"
    data_unit_journal_add = f"{EventScope.data_unit_journal.value}:add"
    data_unit_config_update = f"{EventScope.data_unit_config.value}:update"
    data_unit_data_product_link = f"{EventScope.data_unit_data_product.value}:link"
    data_unit_data_product_unlink = f"{EventScope.data_unit_data_product.value}:unlink"
    data_unit_health_check = f"{EventScope.data_unit_health.value}:check"
    data_unit_profiling_update = f"{EventScope.data_unit_profiling.value}:update"

    spark_finish = f"{EventScope.spark.value}:finish"
    spark_status_update = f"{EventScope.spark_status.value}:update"

    builder_preview = f"{EventScope.builder.value}:preview"

    journal_note_update = f"{EventScope.journal_note.value}:update"
    journal_note_delete = f"{EventScope.journal_note.value}:delete"


class Emitter(abc.ABC):
    def __init__(self, session_id: uuid.UUID | None = None) -> None:
        self.session_id = session_id

    @abc.abstractmethod
    async def emit(self, event_packet: EventPacket) -> None:
        pass


class SyncEmitter(abc.ABC):
    @abc.abstractmethod
    def emit(self, event_packet: EventPacket) -> None:
        pass


class TmpEmitter(Emitter):
    def __init__(self, max_emit_log_line_length: int = MAX_EMIT_LOG_LINE_LENGTH) -> None:
        self.max_emit_log_line_length = max_emit_log_line_length
        self.root_dir = self.ensure_root_folder_exists()

    @staticmethod
    def split_uuid_to_dirs_and_filename(span_id: uuid.UUID) -> tuple[str, str, str]:
        """Split span_id (uuid) into nested directories.

        Helper function that splits span_id (uuid) into nested directories.

        span_dir  span_subdir  span_filename
           87   /      39    / 0265-e23d-4574-b71e-dd1d61a74496

        for the purpose of storing large amount of files and maintaining performance
        with filesystems and other os tools

        Args:
        ----
        span_id: uuid.uuid4

        Returns: tuple (span_dir, span_subdir, span_filename)
        """
        span_id_str = str(span_id)

        span_dir = span_id_str[0:2]
        span_subdir = span_id_str[2:4]
        span_filename = span_id_str[4:]

        return span_dir, span_subdir, span_filename

    @staticmethod
    def ensure_root_folder_exists() -> Path:
        root_dir = Path(ROOT_DIR)
        root_dir.mkdir(exist_ok=True)

        return root_dir

    def get_event_filepath(self, span_id: uuid.UUID, session_id: uuid.UUID | None = None) -> Path:
        """Generate the file path for the emitted event.

        Args:
            span_id (uuid.UUID): The UUID for the event span.
            session_id (typing.Optional[str], optional): The user session ID,
            if applicable. Defaults to None.

        Returns:
            Path: The file path where the emitted event is saved.
        """
        span_dir, span_subdir, span_filename = self.split_uuid_to_dirs_and_filename(span_id)
        path = self.root_dir / span_dir / span_subdir
        if session_id:
            return path / str(session_id) / span_filename
        return path / span_filename

    def read_event_packets(self, span_id: uuid.UUID, session_id: uuid.UUID | None = None) -> EventPackets:
        """Read event packets from a given file path.

        Args:
            span_id (uuid.UUID): The UUID for the event span.
            session_id (typing.Optional[str], optional): The user session ID, if applicable.
            Defaults to None.

        Returns:
            EventPackets: The event packets read from the file.
        """
        event_filepath = self.get_event_filepath(span_id, session_id)

        if event_filepath.is_file():
            with event_filepath.open("r") as f:
                packets = json.load(f)
                event_packets = EventPackets(**packets)
        else:
            event_packets = EventPackets(packets=[])

        return event_packets

    def save_event_packets(
        self,
        span_id: UUID4,
        event_packets: EventPackets,
        session_id: UUID4 | None = None,
    ) -> None:
        """Save event packets with the given span ID and optional session ID.

        Args:
            span_id (UUID4): Unique identifier for the span of the event.
            event_packets (EventPackets): The event packets to be saved.
            session_id (UUID4 | None, optional): Optional session ID. Defaults to None.
        """
        event_filepath = self.get_event_filepath(span_id, session_id)

        event_filepath.parent.mkdir(parents=True, exist_ok=True)
        with event_filepath.open("w") as f:
            json.dump(event_packets.model_dump(), f)

    def clear_oldest_span_id_event(self, span_id: UUID4, session_id: UUID4 | None = None) -> None:
        # Remove span_id file
        dir_0, dir_1, filename = self.split_uuid_to_dirs_and_filename(span_id)

        # empty oldest info in span file
        event_packets = self.read_event_packets(span_id, session_id)
        if event_packets.packets:
            event_packets.packets.pop(0)

        if not event_packets.packets:
            # remove file
            self.get_event_filepath(span_id, session_id).unlink(missing_ok=True)

            # Remove span subdir if empty
            span_sub_dir = self.root_dir / dir_0 / dir_1
            file_count = len([child for child in span_sub_dir.iterdir() if child.is_file()])
            if file_count == 0:
                span_sub_dir.rmdir()

            # Remove span dir if empty
            span_dir = self.root_dir / dir_0
            dir_count = len([child for child in span_dir.iterdir() if child.is_dir()])
            if dir_count == 0:
                span_dir.rmdir()
        else:
            self.save_event_packets(span_id, event_packets)

    def ensure_emit_log_keep_size(self) -> None:
        history_log = self.root_dir / EMIT_LOG_FILENAME

        random_str = "".join(random.choice(string.ascii_lowercase) for i in range(8))  # noqa: S311
        history_log_tmp = self.root_dir / f"{random_str}_{EMIT_LOG_FILENAME}.tmp"

        if history_log.is_file():
            with history_log.open("r") as f:
                number_of_lines = len(f.readlines())

            diff = number_of_lines - (self.max_emit_log_line_length - 1)

            if diff > 0:
                with history_log.open("r") as original_f:
                    # skip the lines
                    for _i in range(diff):
                        # Remove line from event_log by moving the file pointer
                        line = original_f.readline().rstrip("\n")

                        span_id = uuid.UUID(line.split(" ")[-1])
                        self.clear_oldest_span_id_event(span_id)

                    # copy the rest
                    with history_log_tmp.open("w") as tmp_f:
                        for line in original_f:
                            tmp_f.write(line)

                # remove original and rename tmp file
                history_log.unlink()
                history_log_tmp.rename(history_log)

    def update_emit_log(self, event_packet: EventPacket) -> None:
        history_log = self.root_dir / EMIT_LOG_FILENAME

        self.ensure_emit_log_keep_size()

        with history_log.open("a") as f:
            f.write(
                f"{datetime.datetime.fromtimestamp(event_packet.timestamp / 1000.0, tz=datetime.timezone.utc)} {event_packet.message_type} {event_packet.span_id!s}\n",
            )

    async def emit(self, event_packet: EventPacket) -> None:
        try:
            dir_0, dir_1, filename = self.split_uuid_to_dirs_and_filename(event_packet.span_id)
            directory = self.root_dir / dir_0 / dir_1
            directory.mkdir(parents=True, exist_ok=True)

            event_packets = self.read_event_packets(event_packet.span_id, event_packet.session_id)
            event_packets.packets.append(event_packet)

            self.save_event_packets(event_packet.span_id, event_packets, event_packet.session_id)

            self.update_emit_log(event_packet)
        except Exception:
            logger.exception("Failed to emit event.")


def _prepare_event_packet(
    message: str | typing.Any,  # noqa: ANN401
    source: str,
    version: EventVersion,
    span_id: UUID4 | None = None,
    message_type: str = "unknown",
    session_id: UUID4 | None = None,
) -> EventPacket:
    span_id = span_id or uuid.uuid4()

    try:
        deserialized_message = json.loads(message)
    except (json.JSONDecodeError, TypeError):
        deserialized_message = message

    return EventPacket(
        source=source,
        timestamp=int(time.time() * 1000),  # timestamp in ms
        span_id=span_id,
        version=str(version),
        message=deserialized_message,
        message_type=message_type,
        session_id=session_id,
    )


class Event(abc.ABC):
    def __init__(self, emitter: Emitter, source: str = "source", session_id: UUID4 | None = None) -> None:
        self.emitter = emitter
        self.source = source
        self.session_id = session_id

    @abc.abstractmethod
    def version(self) -> EventVersion: ...

    async def emit(
        self,
        message: str,
        span_id: UUID4 | None = None,
        message_type: str = "unknown",
        session_id: UUID4 | None = None,
    ) -> uuid.UUID:
        span_id = span_id or uuid.uuid4()

        event_packet = _prepare_event_packet(
            message=message,
            source=self.source,
            span_id=span_id,
            version=self.version(),
            message_type=message_type,
            session_id=session_id,
        )

        await self.emitter.emit(event_packet)

        return span_id


class SyncEvent(abc.ABC):
    @abc.abstractmethod
    def version(self) -> EventVersion: ...

    def __init__(self, emitter: SyncEmitter, source: str = "source") -> None:
        self.emitter = emitter
        self.source = source

    def emit(
        self,
        message: str,
        span_id: UUID4 | None = None,
        message_type: str = "unknown",
    ) -> UUID4:
        span_id = span_id or uuid.uuid4()

        event_packet = _prepare_event_packet(
            message=message,
            source=self.source,
            span_id=span_id,
            version=self.version(),
            message_type=message_type,
        )

        self.emitter.emit(event_packet)

        return span_id


class ValidationResult(pydantic.BaseModel):
    """Data Quality validation results model."""

    id_: uuid.UUID = pydantic.Field(alias="id")
    entity_identifier: uuid.UUID
    expectations_id: uuid.UUID
    success: bool
    success_percentage: float
    score: int
    column_score: dict[str, int]
    category_weights: dict
    threshold: float
    field_thresholds: dict
    raw_result: dict
    global_stats: dict
    table_stats: dict
    column_stats: dict


class ProfilingResult(pydantic.BaseModel):
    """Profiling results model."""

    id_: uuid.UUID = pydantic.Field(alias="id")
    created_at: datetime.datetime
    entity_identifier: uuid.UUID
    dataset_type: str
    row_count: int
    column_count: int
    profile_sample: float
    column_stats: list[dict] | dict[str, typing.Any]


class ClassificationResult(pydantic.BaseModel):
    """Data classification results model."""

    id_: uuid.UUID = pydantic.Field(alias="id")
    entity_identifier: uuid.UUID
    dataset_type: str
    config: dict
    column_results: dict



import json
import typing
from unittest import mock

import aws4.key_pair
import httpcore
import httpx
import pytest

from neos_common import error
from neos_common.client import base
from tests.conftest import AsyncMock


class DummyClient(base.NeosClient):
    known_errors: typing.ClassVar[set] = {"E000"}
    service_name = "dummy"

    def __init__(self, token, key_pair, partition) -> None:
        assert token is not None or key_pair is not None

        self._token = token
        self._key_pair = key_pair
        self._partition = partition

    @property
    def token(self):
        return self._token

    @property
    def key_pair(self):
        return self._key_pair

    @property
    def partition(self):
        return self._partition


@pytest.fixture
def client():
    return DummyClient(token="token", key_pair=None, partition="ksa")


@pytest.fixture
def signature_client():
    return DummyClient(token="token", key_pair=aws4.key_pair.KeyPair("access-key", "access-secret"), partition="ksa")


class TestNeosClient:
    @pytest.mark.parametrize(
        ("method", "params", "json_payload", "headers"),
        [
            (base.Method.GET, None, None, None),
            (base.Method.GET, None, None, {"auth": "token"}),
            (base.Method.GET, {"a": "b", "c": "d"}, None, {"auth": "token"}),
            (base.Method.POST, None, None, None),
            (base.Method.POST, None, None, {"auth": "token"}),
            (base.Method.POST, None, {"a": "b", "c": "d"}, {"auth": "token"}),
            (base.Method.PUT, None, None, None),
            (base.Method.PUT, None, None, {"auth": "token"}),
            (base.Method.PUT, None, {"a": "b", "c": "d"}, {"auth": "token"}),
            (base.Method.DELETE, None, None, None),
            (base.Method.DELETE, None, None, {"auth": "token"}),
            (base.Method.DELETE, None, {"a": "b", "c": "d"}, {"auth": "token"}),
        ],
    )
    async def test_request(self, client, httpx_mock, method, params, json_payload, headers):
        url = "http://host"
        query = "?{}".format("&".join([f"{k}={v}" for k, v in params.items()])) if params else ""
        body = f"{json.dumps(json_payload)}".encode() if json_payload else None
        httpx_mock.add_response(
            url=f"{url}{query}",
            match_content=body,
            json={"hello": "world"},
            headers=headers,
        )

        r = await client._request(
            url=url,
            method=method,
            params=params,
            headers=headers,
            json=json_payload,
        )

        assert r.json() == {"hello": "world"}

    @pytest.mark.parametrize(
        ("method", "params", "json_payload", "headers"),
        [
            (base.Method.GET, None, None, None),
            (base.Method.GET, None, None, {"auth": "token"}),
            (base.Method.GET, {"a": "b", "c": "d"}, None, {"auth": "token"}),
            (base.Method.POST, None, None, None),
            (base.Method.POST, None, None, {"auth": "token"}),
            (base.Method.POST, None, {"a": "b", "c": "d"}, {"auth": "token"}),
            (base.Method.PUT, None, None, None),
            (base.Method.PUT, None, None, {"auth": "token"}),
            (base.Method.PUT, None, {"a": "b", "c": "d"}, {"auth": "token"}),
            (base.Method.DELETE, None, None, None),
            (base.Method.DELETE, None, None, {"auth": "token"}),
            (base.Method.DELETE, None, {"a": "b", "c": "d"}, {"auth": "token"}),
        ],
    )
    async def test_signed_request(self, signature_client, httpx_mock, method, params, json_payload, headers):
        url = "http://host"
        query = "?{}".format("&".join([f"{k}={v}" for k, v in params.items()])) if params else ""
        body = f"{json.dumps(json_payload)}".encode() if json_payload else None
        httpx_mock.add_response(
            url=f"{url}{query}",
            match_content=body,
            json={"hello": "world"},
            headers=headers,
        )

        r = await signature_client._request(url, method, params, headers, json_payload)

        assert r.json() == {"hello": "world"}

    async def test_token_request_generates_bearer_auth(self, client, monkeypatch):
        httpx_client = mock.MagicMock()
        mock_client = AsyncMock()
        httpx_client.__aenter__.return_value = mock_client
        monkeypatch.setattr(base.httpx, "AsyncClient", mock.MagicMock(return_value=httpx_client))

        await client._request("url", base.Method.GET)

        assert mock_client.request.call_args == mock.call(
            url="url",
            method="GET",
            params=None,
            json=None,
            headers=None,
            auth=base.NeosBearerClientAuth("token"),
        )

    async def test_empty_token_request_excludes_auth(self, monkeypatch):
        client = DummyClient(token="", key_pair=None, partition="ksa")
        httpx_client = mock.MagicMock()
        mock_client = AsyncMock()
        httpx_client.__aenter__.return_value = mock_client
        monkeypatch.setattr(base.httpx, "AsyncClient", mock.MagicMock(return_value=httpx_client))

        await client._request("url", base.Method.GET)

        assert mock_client.request.call_args == mock.call(
            url="url",
            method="GET",
            params=None,
            json=None,
            headers=None,
            auth=None,
        )

    async def test_signed_request_generates_signature_auth(self, signature_client, monkeypatch):
        httpx_client = mock.MagicMock()
        mock_client = AsyncMock()
        httpx_client.__aenter__.return_value = mock_client
        monkeypatch.setattr(base.httpx, "AsyncClient", mock.MagicMock(return_value=httpx_client))

        await signature_client._request("url", base.Method.GET)

        assert mock_client.request.call_args == mock.call(
            url="url",
            method="GET",
            params=None,
            json=None,
            headers=None,
            auth=base.HttpxAWS4Auth(
                aws4.key_pair.KeyPair("access-key", "access-secret"),
                "service",
                "ksa",
                aws4.AuthSchema("NEOS4-HMAC-SHA256", "x-neos"),
            ),
        )

    async def test_request_handles_timeout(self, client, httpx_mock):
        httpx_mock.add_exception(httpcore.ReadTimeout("Timed out"))
        with pytest.raises(error.ServiceTimeoutError):
            await client._request("url", method=base.Method.GET)

    async def test_request_handles_connect_error(self, client, httpx_mock):
        httpx_mock.add_exception(httpx.ConnectError("Unable to connect"))
        with pytest.raises(error.ServiceConnectionError):
            await client._request("url", method=base.Method.GET)

    async def test_get(self, client, monkeypatch):
        monkeypatch.setattr(client, "_request", AsyncMock(return_value="response"))

        r = await client._get("url", "params", "headers")

        assert r == "response"
        assert client._request.call_args == mock.call(
            url="url",
            method=base.Method.GET,
            params="params",
            headers="headers",
        )

    async def test_post(self, client, monkeypatch):
        monkeypatch.setattr(client, "_request", AsyncMock(return_value="response"))

        r = await client._post("url", "json", "headers")

        assert r == "response"
        assert client._request.call_args == mock.call(
            url="url",
            method=base.Method.POST,
            json="json",
            headers="headers",
        )

    async def test_put(self, client, monkeypatch):
        monkeypatch.setattr(client, "_request", AsyncMock(return_value="response"))

        r = await client._put("url", "json", "headers")

        assert r == "response"
        assert client._request.call_args == mock.call(
            url="url",
            method=base.Method.PUT,
            json="json",
            headers="headers",
        )

    async def test_delete(self, client, monkeypatch):
        monkeypatch.setattr(client, "_request", AsyncMock(return_value="response"))

        r = await client._delete("url", "json", "headers")

        assert r == "response"
        assert client._request.call_args == mock.call(
            url="url",
            method=base.Method.DELETE,
            json="json",
            headers="headers",
        )

    def test_process_unexpected_response(self, client):
        json = {"kong": "error"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=400)

        with pytest.raises(error.UnhandledServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "unhandled-service-api"
        assert e.value.message == "Unhandled dummy api response."
        assert e.value.debug is None
        assert e.value.status == "bad-request"

    def test_process_invalid_json(self, client):
        response = mock.Mock(json=mock.Mock(side_effect=json.JSONDecodeError("Ups", "doc", 1)), status_code=400)

        with pytest.raises(error.UnhandledServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "unhandled-service-api"
        assert e.value.message == "Invalid dummy api JSON format response."
        assert e.value.debug == "Ups: 1"
        assert e.value.status == "bad-request"

    def test_process_rate_limit_timeout(self, client):
        json = {"kong": "error"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=429)

        with pytest.raises(error.UnhandledServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "unhandled-service-api"
        assert e.value.message == "dummy api rate limit error."
        assert e.value.debug is None
        assert e.value.status == "too-many-requests"

    def test_process_response_unknown_error(self, client):
        json = {"type": "1", "title": "message", "details": "debug_message"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=400)

        with pytest.raises(error.UnhandledServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "1"
        assert e.value.message == "Unhandled dummy api error response."
        assert e.value.debug == "debug_message"
        assert e.value.status == "bad-request"

    def test_process_response_validation_error(self, client):
        json = {"type": "1", "title": "message", "errors": {"validation": "error"}}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=422)

        with pytest.raises(error.ServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "1"
        assert e.value.message == "dummy request validation error."
        assert e.value.details == {"validation": "error"}
        assert e.value.status == "unprocessable-entity"

    @pytest.mark.parametrize(
        ("code", "expected"),
        [
            ("E000", error.ServiceApiError),
        ],
    )
    def test_process_response_known_error(self, code, expected, client):
        json = {"type": code, "title": "message", "details": "debug_message"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=400)

        with pytest.raises(expected) as e:
            client.process_response(response)

        assert e.value.debug == "debug_message"



import asyncio
import json
import logging
import socket
import time

from neos_common import error
from neos_common.socket import util

logger = logging.getLogger(__name__)


class AsyncTCPClient:
    """Base implementation for asyncio stream socket service client."""

    def __init__(
        self,
        host: str,
        port: int,
        encoder: type[json.JSONEncoder] | None = None,
    ) -> None:
        """TCPClient instantiator.

        Args:
        ----
        host: socket service host name
        port: socket service port
        encoder: json encoder for request messages
        """
        self.addr = (host, port)
        self.encoder = encoder

    async def send_request(self, request: dict) -> dict:
        """Send a request to socket service.

        Handle socket connection errors, and response timeouts.
        """
        try:
            reader, writer = await asyncio.open_connection(*self.addr)
            async_socket = util.AsyncSocket(reader, writer)
        except OSError as e:
            raise error.ServiceConnectionError(
                message="Unable to connect to service.",
                debug=f"{self.addr}: {e!s}",
            ) from e

        try:
            async_socket.write(util.encode(request, encoder=self.encoder))
            response = await async_socket.read()
        finally:
            async_socket.close()

        return response


class TCPClient:
    """Base implementation for socket service client."""

    def __init__(
        self,
        host: str,
        port: int,
        timeout: int = 10,
        wait: int = 10,
        encoder: type[json.JSONEncoder] | None = None,
    ) -> None:
        """TCPClient instantiator.

        Args:
        ----
        host: socket service host name
        port: socket service port
        timeout: number of seconds to wait for a connection
        wait: number of seconds to wait for a response from service
        encoder: json encoder for request messages
        """
        self.addr = (host, port)
        self.timeout = timeout
        self.wait = wait
        self.encoder = encoder

    def send_request(self, request: dict) -> dict:
        """Send a request to socket service.

        Handle socket connection errors, and response timeouts.
        """
        t0 = time.time()
        try:
            sock = socket.create_connection(
                self.addr,
                timeout=self.timeout,
            )
        except socket.gaierror as e:
            raise error.ServiceConnectionError(
                message="Unable to connect to service.",
                debug=f"{self.addr}: {e!s}",
            ) from e

        try:
            util.send_msg(sock, util.encode(request, encoder=self.encoder))
            t0 = time.time()
            response = None
            while not response:
                try:
                    response = util.decode(util.recv_msg(sock))
                except Exception as e:
                    raise error.ServiceConnectionError(
                        message="Unable to process request to service.",
                        debug=f"{self.addr}: {e.__class__.__name__}({e!s})",
                    ) from e
                if time.time() - t0 > self.wait:
                    msg = "Timed out waiting for resp."
                    raise TimeoutError(msg)
        finally:
            sock.close()
        return response



import json
from unittest import mock

import pytest
from aws4.key_pair import KeyPair

from neos_common import base, error, schema
from neos_common.authorization import token
from neos_common.client.hub_client import HubClient
from neos_common.schema import Contract, ContractSubscription, ContractVisibility


@pytest.fixture
def client():
    return HubClient(
        "https://hub-host",
        token.TokenData("user-id", "auth_token", []),
        key_pair=None,
        account="root",
        partition="ksa",
    )


@pytest.fixture
def signature_client():
    return HubClient(
        "https://hub-host",
        None,
        key_pair=KeyPair("access_key", "access_secret"),
        account="root",
        partition="ksa",
    )


class TestHubClient:
    def test_init_bearer(self):
        hub_client = HubClient("host", "token", None, "account", "partition")
        assert hub_client._host == "host"
        assert hub_client._token == "token"
        assert hub_client._key_pair is None
        assert hub_client._account == "account"
        assert hub_client._partition == "partition"

    def test_init_sign(self):
        hub_client = HubClient("host", None, "key_pair", "account", "partition")
        assert hub_client._host == "host"
        assert hub_client._token is None
        assert hub_client._key_pair == "key_pair"
        assert hub_client._account == "account"
        assert hub_client._partition == "partition"

    def test_validate_token_url(self, client):
        assert client.validate_token_url == "https://hub-host/iam/validate/token"

    def test_validate_signature_url(self, client):
        assert client.validate_signature_url == "https://hub-host/iam/validate/signature"

    @pytest.mark.parametrize(
        ("code", "expected_error"),
        [
            ("authorization-required", error.AuthorizationRequiredError),
            ("invalid-authorization", error.InvalidAuthorizationError),
            ("insufficient-permissions", error.InsufficientPermissionsError),
            ("invalid-resource-format", error.InvalidResourceFormatError),
            ("identity-access-manager-error", error.IdentityAccessManagerError),
            ("server-error", error.ServiceApiError),
        ],
    )
    def test_process_response_mapping_error(self, client, code, expected_error):
        json = {"type": code, "title": "message", "details": "debug message"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=422)

        with pytest.raises(expected_error):
            client.process_response_with_mapping(response)

    def test_process_response_unknown_error(self, client):
        json = {"type": "1", "title": "message", "details": "debug_message"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=400)

        with pytest.raises(error.UnhandledServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "1"
        assert e.value.message == "Unhandled Hub api error response."
        assert e.value.debug == "debug_message"
        assert e.value.status == "bad-request"

    def test_process_response_validation_error(self, client):
        json = {"type": "1", "title": "message", "errors": {"validation": "error"}}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=422)

        with pytest.raises(error.ServiceApiError) as e:
            client.process_response(response)

        assert e.value.reason == "1"
        assert e.value.message == "Hub request validation error."
        assert e.value.details == {"validation": "error"}
        assert e.value.status == "unprocessable-entity"

    @pytest.mark.parametrize(
        ("code", "expected"),
        [
            ("policy-not-found", error.ServiceApiError),
            ("core-already-registered", error.ServiceApiError),
        ],
    )
    def test_process_response_known_error(self, code, expected, client):
        json = {"type": code, "title": "message", "details": "debug_message"}
        response = mock.Mock(json=mock.Mock(return_value=json), status_code=400)

        with pytest.raises(expected) as e:
            client.process_response(response)

        assert e.value.debug == "debug_message"

    async def test_validate_token_with_return_resource(self, httpx_mock, client):
        resources_payload = {
            "principal_id": "user_id",
            "resources": ["nrn:ksa:core:sc:root:product:my-product"],
        }
        httpx_mock.add_response(
            url="https://hub-host/iam/validate/token?principal_id=user_id&action=core:announce&resource=nrn%3Aksa%3Acore%3Asc%3Aroot%3Aproduct%3Amy-product&return_allowed_resources=true&logic_operator=and",
            json=resources_payload,
        )

        user_id, resources = await client.validate_token(
            principal="user_id",
            actions=[base.Action.core_announce],
            resources=["nrn:ksa:core:sc:root:product:my-product"],
            logic_operator="and",
            return_allowed_resources=True,
        )
        assert resources == resources_payload["resources"]
        assert user_id == "user_id"

    async def test_validate_token_without_return_resource(self, httpx_mock, client):
        resources_payload = {
            "principal_id": "user_id",
            "resources": [],
        }
        httpx_mock.add_response(
            url="https://hub-host/iam/validate/token?principal_id=user_id&action=core:announce&resource=nrn%3Aksa%3Acore%3Asc%3Aroot%3Aproduct%3Amy-product&return_allowed_resources=false&logic_operator=or",
            json=resources_payload,
        )

        user_id, resources = await client.validate_token(
            principal="user_id",
            actions=[base.Action.core_announce],
            resources=["nrn:ksa:core:sc:root:product:my-product"],
            logic_operator="or",
            return_allowed_resources=False,
        )
        assert resources == resources_payload["resources"]
        assert user_id == "user_id"

    async def test_validate_signature(self, httpx_mock, client):
        resources_payload = {
            "principal_id": "user_id",
            "resources": ["nrn:ksa:core:sc:root:product:my-product"],
        }
        httpx_mock.add_response(
            url="https://hub-host/iam/validate/signature?access_key_id=access-key-id&auth_schema=NEOS4&scope=scope&challenge=challenge&signed_challenge=signed-challenge&action=core:announce&resource=nrn%3Aksa%3Acore%3Asc%3Aroot%3Aproduct%3Amy-product&return_allowed_resources=true&logic_operator=and",
            json=resources_payload,
        )

        user_id, resources = await client.validate_signature(
            access_key="access-key-id",
            auth_schema="NEOS4",
            scope="scope",
            challenge="challenge",
            signed_challenge="signed-challenge",
            actions=[base.Action.core_announce],
            resources=["nrn:ksa:core:sc:root:product:my-product"],
            logic_operator="and",
            return_allowed_resources=True,
        )
        assert resources == resources_payload["resources"]
        assert user_id == "user_id"

    async def test_get_allowed_resources(self, httpx_mock, client):
        httpx_mock.add_response(
            url="https://hub-host/iam/principal/user-id/resource?action=product:consume",
            json={
                "resources": ["a", "b"],
                "principal_id": "user-id",
            },
        )

        resources = await client.get_allowed_resources("user-id", base.Action.product_consume)
        assert resources == ["a", "b"]

    async def test_announce_core(self, httpx_mock, client):
        httpx_mock.add_response(
            method="POST",
            url="https://hub-host/registry/core/identifier/announce",
            headers={"X-Access-Key": "core-access-key"},
            match_content=b'{"host": "host", "version": "v1.2.3"}',
            json={},
        )

        await client.announce_core("host", version="v1.2.3", core_identifier="identifier")

    async def test_register_data_product_approval_false(self, httpx_mock, client):
        # Set up expected contract with approval set to False
        expected_contract = Contract(
            visibility=ContractVisibility.PUBLIC,
            subscription=ContractSubscription(approval=False),
        )

        # Mock the response to simulate the API's response when approval is False
        httpx_mock.add_response(
            method="POST",
            url="https://hub-host/registry/core/core-identifier/data_product",
            headers={"X-Access-Key": "core-access-key"},
            json={
                "urn": "urn",
                "name": "name",
                "metadata": {"hello": "world"},
                "description": "description",
                "label": "ABC",
                "sanitized_name": "sanitized_name",
                "data_product_type": "data_product_type",
                "engine": "engine",
                "table": "table",
                "identifier": "identifier",
                "contract": {"visibility": "public", "subscription": {"approval": False}},
            },
        )

        # Execute the method with approval set to False
        await client.register_data_product(
            "urn",
            "name",
            metadata={"hello": "world"},
            description="description",
            label="ABC",
            sanitized_name="sanitized_name",
            data_product_type="data_product_type",
            engine="engine",
            table="table",
            identifier="identifier",
            core_identifier="core-identifier",
            contract=expected_contract,
        )

        request = httpx_mock.get_request()
        request_data = json.loads(request.content)

        # Assert that approval is set to False in the contract
        assert request_data["contract"]["subscription"]["approval"] is False

    async def test_register_data_product(self, httpx_mock, client):
        httpx_mock.add_response(
            method="POST",
            url="https://hub-host/registry/core/core-identifier/data_product",
            headers={"X-Access-Key": "core-access-key"},
            match_content=bytes(
                json.dumps({
                    "urn": "urn",
                    "name": "name",
                    "metadata": {"hello": "world"},
                    "description": "description",
                    "label": "ABC",
                    "sanitized_name": "sanitized_name",
                    "data_product_type": "data_product_type",
                    "engine": "engine",
                    "table": "table",
                    "identifier": "identifier",
                    "contract": {"visibility": "public", "subscription": {"approval": True}},
                }),
                "utf-8",
            ),
            json={},
        )

        await client.register_data_product(
            "urn",
            "name",
            metadata={"hello": "world"},
            description="description",
            label="ABC",
            sanitized_name="sanitized_name",
            data_product_type="data_product_type",
            engine="engine",
            table="table",
            identifier="identifier",
            core_identifier="core-identifier",
            contract=schema.Contract(visibility="public", subscription=schema.ContractSubscription(approval=True)),
        )

    async def test_register_private_data_product(self, httpx_mock, client):
        httpx_mock.add_response(
            method="POST",
            url="https://hub-host/registry/core/core-identifier/data_product",
            headers={"X-Access-Key": "core-access-key"},
            match_content=bytes(
                json.dumps({
                    "urn": "urn",
                    "name": "name",
                    "metadata": {"hello": "world"},
                    "description": "description",
                    "label": "ABC",
                    "sanitized_name": "sanitized_name",
                    "data_product_type": "data_product_type",
                    "engine": "engine",
                    "table": "table",
                    "identifier": "identifier",
                    "contract": {"visibility": "private", "subscription": {"approval": True}},
                }),
                "utf-8",
            ),
            json={},
        )

        await client.register_data_product(
            "urn",
            "name",
            metadata={"hello": "world"},
            description="description",
            label="ABC",
            sanitized_name="sanitized_name",
            data_product_type="data_product_type",
            engine="engine",
            table="table",
            identifier="identifier",
            core_identifier="core-identifier",
            contract=schema.Contract(visibility="private", subscription=schema.ContractSubscription(approval=True)),
        )

    async def test_update_data_product_metadata(self, httpx_mock, client):
        httpx_mock.add_response(
            method="POST",
            url="https://hub-host/registry/core/core-identifier/data_product/metadata",
            headers={"X-Access-Key": "core-access-key"},
            match_content=b'{"urn": "urn", "metadata": {"hello": "world"}, "description": "desc"}',
            json={},
        )

        await client.update_data_product_metadata(
            "urn",
            metadata={"hello": "world"},
            description="desc",
            core_identifier="core-identifier",
        )

    async def test_deregister_data_product(self, httpx_mock, client):
        httpx_mock.add_response(
            method="DELETE",
            url="https://hub-host/registry/core/core-identifier/data_product",
            headers={"X-Access-Key": "core-access-key"},
            match_content=b'{"urn": "urn"}',
            json={},
        )

        await client.deregister_data_product("urn", core_identifier="core-identifier")



from unittest import mock

import neos_common.config
from neos_common.config import Config


class TestConfig:
    def test_load(self, monkeypatch):
        monkeypatch.setenv("NC_POSTGRES_HOST", "host")
        monkeypatch.setenv("NC_POSTGRES_USER", "user")
        monkeypatch.setenv("NC_POSTGRES_DATABASE", "database")
        monkeypatch.setenv("NC_POSTGRES_PASSWORD", "pword")
        monkeypatch.setattr(neos_common.config.dotenv, "find_dotenv", mock.Mock(return_value=""))

        Config.load()

        assert neos_common.config.dotenv.find_dotenv.call_args == mock.call(usecwd=True)

    def test_postgres_dsn(self, monkeypatch):
        monkeypatch.setenv("NC_POSTGRES_HOST", "host")
        monkeypatch.setenv("NC_POSTGRES_USER", "user")
        monkeypatch.setenv("NC_POSTGRES_DATABASE", "database")
        monkeypatch.setenv("NC_POSTGRES_PASSWORD", "pword")

        config = Config()

        dsn = config.postgres_dsn

        assert dsn == "postgresql://user:pword@host:5432/database"

    def test_logging_config(self, monkeypatch):
        monkeypatch.setenv("NC_POSTGRES_HOST", "host")
        monkeypatch.setenv("NC_POSTGRES_USER", "user")
        monkeypatch.setenv("NC_POSTGRES_DATABASE", "database")
        monkeypatch.setenv("NC_POSTGRES_PASSWORD", "pword")

        config = Config()

        assert config.logging_config == {
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                    "format": "[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                },
            },
            "handlers": {
                "default": {
                    "class": "logging.StreamHandler",
                    "formatter": "default",
                    "level": "INFO",
                },
            },
            "loggers": {
                "common": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "pogo_migrate": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "neos_common": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
            },
            "version": 1,
        }



import typing
import uuid
from dataclasses import dataclass

from neos_common.authorization.token import TokenData
from neos_common.base import Action, ResourceLike, ResourceType
from neos_common.config import ApiConfig, Config

if typing.TYPE_CHECKING:
    import fastapi

    from neos_common.client.hub_client import HubClient
    from neos_common.client.keycloak_client import KeycloakClient


class SignatureValidator(typing.Protocol):
    """Define the base requirements for an object that can validate signed requests."""

    async def validate(
        self,
        request: "fastapi.Request",
        action: list[Action | str],
        resource: list[ResourceLike],
        logic_operator: str,
        *,
        return_allowed_resources: bool,
    ) -> tuple[uuid.UUID, list[str]]:
        """Validate a request and return the associated user_id and resources."""
        ...  # pragma: no cover


class AccessValidator:
    async def validate(
        self,
        user_id: uuid.UUID,
        actions: list[Action | str],
        resources: list[ResourceLike],
        logic_operator: str,
        *,
        return_allowed_resources: bool = True,
    ) -> tuple[uuid.UUID, list[str]]: ...  # pragma: no cover


class ConfigDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns Config."""

    async def __call__(self, request: "fastapi.Request") -> Config: ...  # pragma: no cover


class SignatureValidatorDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns a SignatureValidator."""

    async def __call__(self) -> SignatureValidator: ...  # pragma: no cover


class HubClientDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns an HubClient."""

    async def __call__(self, config: Config) -> "HubClient": ...  # pragma: no cover


class AccessValidatorDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns an AccessValidator."""

    async def __call__(self, hub_client: "HubClient", token: TokenData) -> AccessValidator: ...  # pragma: no cover


class KeycloakClientDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns a KeycloakClient."""

    async def __call__(self, config: ApiConfig) -> "KeycloakClient": ...  # pragma: no cover


@dataclass
class ActionResource:
    action: Action
    resource_type: ResourceType
    resource: ResourceLike
    resource_extractor: "typing.Callable[[fastapi.Request], dict[str, str]] | None" = None


@dataclass
class ActionResourceList:
    action_resources: list[ActionResource]
    logic_operator: typing.Literal["and", "or"] = "and"


class AuthorizationDependency(typing.Protocol):
    """Define the base requirements for a dependency that validates authorization."""

    def __init__(
        self,
        action_resources: ActionResourceList | list[ActionResource] | ActionResource,
        *,
        return_allowed_resources: bool = False,
    ) -> None: ...  # pragma: no cover

    async def __call__(
        self,
        request: "fastapi.Request",
        keycloak_client: "KeycloakClient",
        config: Config,
        signature_validator: SignatureValidator,
        access_validator: AccessValidator,
    ) -> TokenData: ...  # pragma: no cover



from __future__ import annotations

import dataclasses
import logging
import smtplib
import typing
from dataclasses import dataclass
from email.mime.text import MIMEText

import jinja2

if typing.TYPE_CHECKING:
    import pathlib


logger = logging.getLogger(__name__)


@dataclass
class TransactionalEmail:
    to_email: str
    to_name: str
    from_email: str
    from_name: str

    subject: typing.ClassVar[str]
    template_name: typing.ClassVar[str]
    template_path: typing.ClassVar[pathlib.Path]

    @property
    def fields(self: typing.Self) -> dict[str, str]:
        return {
            field.name: str(
                getattr(
                    self,
                    field.name,
                ),
            )
            for field in dataclasses.fields(self)
        }

    @property
    def message(self: typing.Self) -> dict[str, typing.Any]:
        environment = jinja2.Environment(  # noqa: S701
            loader=jinja2.FileSystemLoader(self.template_path),
        )
        template = environment.get_template(self.template_name + ".j2")

        fields = self.fields
        return template.render(**fields)


@dataclasses.dataclass
class SMTPConfig:
    smtp_server: str
    smtp_port: int
    smtp_username: str
    smtp_password: str


class SMTPEmailHandler:
    def __init__(
        self: typing.Self,
        *,
        config: SMTPConfig,
        from_email: str,
        from_name: str,
    ) -> None:
        self.from_email = from_email
        self.from_name = from_name
        self.smtp_server = config.smtp_server
        self.smtp_port = config.smtp_port
        self.smtp_username = config.smtp_username
        self.smtp_password = config.smtp_password

    @property
    def client(self: typing.Self) -> smtplib.SMTP:
        client = smtplib.SMTP(self.smtp_server, self.smtp_port)
        client.login(self.smtp_username, self.smtp_password)
        return client

    def setup(self) -> None:
        self._template_classes = {cls.__name__: cls for cls in TransactionalEmail.__subclasses__()}

    def __getitem__(self: typing.Self, name: str) -> TransactionalEmail:
        """Fetch an email template by name."""
        return self._template_classes[name]

    def send(self: typing.Self, template_cls: type[TransactionalEmail], **template_kwargs: str) -> None:
        template = template_cls(
            from_email=self.from_email,
            from_name=self.from_name,
            **template_kwargs,
        )

        msg = MIMEText(template.message)
        msg["From"] = self.from_email
        msg["To"] = template.to_email
        msg["Subject"] = template.subject

        try:
            self.client.sendmail(
                from_addr=self.from_email,
                to_addrs=[template.to_email],
                msg=msg.as_string(),
            )
        except Exception:
            logger.exception("Error occurred sending email message.")



from neos_common.base import ResourceBase


class Resource(ResourceBase):
    @classmethod
    def get_resource_id_template(cls):
        return "{resource_id}"

    @classmethod
    def format_resource_id(cls, *args):
        return cls.get_resource_id_template().format(resource_id=args[0])



from unittest import mock

import fastapi
import pytest
from starlette import testclient

from neos_common.api import get_error_codes, get_permissions
from neos_common.authorization.util import create_openapi_info
from neos_common.base import Action, ResourceBase, ResourceType


class Resource(ResourceBase):
    @classmethod
    def get_resource_id_template(cls):
        return "{resource_id}"

    @classmethod
    def format_resource_id(cls, *args):
        return cls.get_resource_id_template().format(resource_id=args[0])

    @staticmethod
    def generate_from_config(
        config,  # noqa: ARG004
        resource_type,
        resource_id,
        partition,
        account_id,
    ) -> "Resource":
        return Resource.generate(
            xrn="urn",
            partition=partition,
            service="core",
            identifier="",
            account_id=account_id,
            resource_type=resource_type.value,
            resource_id=str(resource_id) if resource_id else None,
        )


@pytest.fixture
def client():
    app = fastapi.FastAPI()
    config = mock.Mock()

    def get_something() -> None:
        pass

    def get_something_else() -> None:
        pass

    app.get(
        "/something",
        summary="Get Something.",
        openapi_extra=create_openapi_info(
            config,
            Resource,
            (Action.core_register, ResourceType.core, None, "{partition}", "root"),
        ),
    )(get_something)
    app.get(
        "/something_else",
        summary="Get Something Else.",
        openapi_extra=create_openapi_info(
            config,
            Resource,
            (Action.core_register, ResourceType.core, None, "{partition}", "root"),
            (Action.core_access, ResourceType.core, None, "{partition}", "root"),
            logic_operator="or",
        ),
    )(get_something_else)

    return testclient.TestClient(app)


def test_get_error_codes():
    assert get_error_codes("tests.error_codes_fixture").model_dump() == {
        "errors": [
            {
                "class_name": "NoCodeError",
                "type": "no-code",
                "title": "Test error.",
            },
            {
                "class_name": "TestError",
                "type": "test-error",
                "title": "Test error.",
            },
            {
                "class_name": "UnauthorisedError",
                "type": "unauthorised",
                "title": "",
            },
        ],
    }


def test_get_error_codes_list():
    assert get_error_codes(["tests.error_codes_fixture"]).model_dump() == {
        "errors": [
            {
                "class_name": "NoCodeError",
                "type": "no-code",
                "title": "Test error.",
            },
            {
                "class_name": "TestError",
                "type": "test-error",
                "title": "Test error.",
            },
            {
                "class_name": "UnauthorisedError",
                "type": "unauthorised",
                "title": "",
            },
        ],
    }


def test_get_permissions(client):
    assert get_permissions(
        client.app,
        ignore_routes=["/openapi.json", "/docs/oauth2-redirect", "/docs", "/redoc"],
    ).model_dump() == {
        "routes": [
            {
                "logic_operator": "and",
                "methods": "GET",
                "path": "/something",
                "permission_pairs": [{"action": "core:register", "resource": "urn:{partition}:core::root:core"}],
                "summary": "Get Something.",
            },
            {
                "logic_operator": "or",
                "methods": "GET",
                "path": "/something_else",
                "permission_pairs": [
                    {"action": "core:register", "resource": "urn:{partition}:core::root:core"},
                    {"action": "core:access", "resource": "urn:{partition}:core::root:core"},
                ],
                "summary": "Get Something Else.",
            },
        ],
    }



import os
import pathlib
from unittest import mock

import pytest

from neos_common import changelog


@pytest.fixture
def cwd(tmp_path):
    orig = pathlib.Path.cwd()

    try:
        os.chdir(str(tmp_path))
        yield tmp_path
    finally:
        os.chdir(orig)


def test_generate_docs(cwd):
    (cwd / "docs").mkdir()
    ctx = mock.Mock(config=mock.Mock(custom={"pdoc": {"modules": ["neos_common"]}}))
    paths = changelog.generate_docs(ctx, "0.0.0")

    assert paths == [
        "docs/neos_common/index.md",
        "docs/neos_common/api.md",
        "docs/neos_common/authorization/index.md",
        "docs/neos_common/authorization/base.md",
        "docs/neos_common/authorization/token.md",
        "docs/neos_common/authorization/util.md",
        "docs/neos_common/authorization/validator.md",
        "docs/neos_common/base.md",
        "docs/neos_common/changelog.md",
        "docs/neos_common/cli.md",
        "docs/neos_common/client/index.md",
        "docs/neos_common/client/base.md",
        "docs/neos_common/client/email.md",
        "docs/neos_common/client/hub_client.md",
        "docs/neos_common/client/kafka_client.md",
        "docs/neos_common/client/keycloak_client.md",
        "docs/neos_common/config.md",
        "docs/neos_common/error.md",
        "docs/neos_common/event.md",
        "docs/neos_common/middleware/index.md",
        "docs/neos_common/middleware/timing.md",
        "docs/neos_common/schema.md",
        "docs/neos_common/socket/index.md",
        "docs/neos_common/socket/client.md",
        "docs/neos_common/socket/server.md",
        "docs/neos_common/socket/util.md",
        "docs/neos_common/util.md",
    ]



from unittest import mock

import pytest

from neos_common.socket import util
from tests.conftest import AsyncMock

RAW_MESSAGE = b'{"hello": "world"}'
MESSAGE_HEADER = b"\x00\x00\x00\x12"


@pytest.mark.parametrize(
    ("obj", "expected"),
    [
        ({"hello": "world"}, b'{"hello": "world"}'),
    ],
)
def test_encode(obj, expected):
    r = util.encode(obj)

    assert r == expected


class UnhandledJsonObject:
    pass


def test_encode_fails_for_unhandled_object():
    with pytest.raises(TypeError):
        util.encode({"hello": UnhandledJsonObject()})


def test_decode():
    r = util.decode(RAW_MESSAGE)

    assert r == {"hello": "world"}


def test_decode_no_message():
    r = util.decode(None)

    assert r is None


def test_len_inbytes():
    assert util.len_inbytes(RAW_MESSAGE) == MESSAGE_HEADER


def test_add_msg_header():
    assert util.add_msg_header(RAW_MESSAGE) == b'\x00\x00\x00\x12{"hello": "world"}'


def test_len_frombytes():
    assert util.len_frombytes(MESSAGE_HEADER) == 18


def test_recvall():
    sock = mock.Mock()
    sock.recv.side_effect = [
        b'{"hel',  # pragma: no-spell-check
        b'lo": ',
        b'"worl',  # pragma: no-spell-check
        b'd"}',
    ]

    data = util.recvall(sock, 18)

    assert data == RAW_MESSAGE
    assert sock.recv.call_args_list == [
        mock.call(18),
        mock.call(13),
        mock.call(8),
        mock.call(3),
    ]


def test_recvall_no_data():
    sock = mock.Mock()
    sock.recv.return_value = b""

    data = util.recvall(sock, 18)

    assert data is None
    assert sock.recv.call_args == mock.call(18)


def test_recv_msg():
    sock = mock.Mock()
    sock.recv.side_effect = [
        MESSAGE_HEADER,
        b'{"hello": "world"}',
    ]

    data = util.recv_msg(sock)

    assert data == RAW_MESSAGE
    assert sock.recv.call_args_list == [
        mock.call(4),
        mock.call(18),
    ]


def test_recv_msg_no_data():
    sock = mock.Mock()
    sock.recv.side_effect = [
        b"",
    ]

    data = util.recv_msg(sock)

    assert data is None
    assert sock.recv.call_args_list == [
        mock.call(4),
    ]


def test_send_msg():
    sock = mock.Mock()

    util.send_msg(sock, RAW_MESSAGE)

    assert sock.sendall.call_args == mock.call(
        MESSAGE_HEADER + RAW_MESSAGE,
    )


class TestAsyncSocket:
    def test_write(self):
        s = util.AsyncSocket(AsyncMock(), mock.Mock())

        s.write(b"message")

        assert s.writer.write.call_args == mock.call(b"\x00\x00\x00\x07message")

    async def test_read(self):
        s = util.AsyncSocket(AsyncMock(), mock.Mock())
        s.reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"hello": "world"}',
        ]

        d = await s.read()

        assert d == {"hello": "world"}

    async def test_read_all(self):
        s = util.AsyncSocket(AsyncMock(), mock.Mock())
        s.reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"hello": "w',
            b'orld"}',
        ]

        d = await s.read()

        assert d == {"hello": "world"}



from unittest import mock

import pytest

keycloak = pytest.importorskip("keycloak")
keycloak_client = pytest.importorskip("neos_common.client.keycloak_client")


@pytest.fixture
def client():
    return keycloak_client.KeycloakClient(
        "https://keycloak-host",
        "test-realm",
        "test-client",
        "test-secret",
    )


class TestKeycloakClient:
    def test_init(self):
        c = keycloak_client.KeycloakClient(
            "https://keycloak-host",
            "test-realm",
            "test-client",
            "test-secret",
        )

        assert c._host == "https://keycloak-host"
        assert c._realm == "test-realm"
        assert c._client_id == "test-client"
        assert c._client_secret == "test-secret"

    def test_openid(self, client, monkeypatch):
        mocked_thing = mock.Mock()
        monkeypatch.setattr(keycloak_client.keycloak, "KeycloakOpenID", mocked_thing)
        client.openid  # noqa: B018
        mocked_thing.assert_called_once_with(
            server_url="https://keycloak-host/auth",
            client_id="test-client",
            realm_name="test-realm",
            client_secret_key="test-secret",
        )

    def test_introspect_token(self, client, monkeypatch):
        monkeypatch.setattr(keycloak_client.keycloak.KeycloakOpenID, "introspect", mock.Mock())

        client.introspect_token("token")

        keycloak_client.keycloak.KeycloakOpenID.introspect.assert_called_once_with("token")

    def test_token_by_password(self, client, monkeypatch):
        mocked_thing = mock.Mock()
        monkeypatch.setattr(keycloak_client.keycloak.KeycloakOpenID, "token", mocked_thing)
        client.get_token_by_password("user", "password")
        mocked_thing.assert_called_once_with("user", "password")

    def test_refresh_token(self, client, monkeypatch):
        mocked_thing = mock.Mock()
        monkeypatch.setattr(keycloak_client.keycloak.KeycloakOpenID, "refresh_token", mocked_thing)
        client.refresh_token("token")
        mocked_thing.assert_called_once_with("token")

    def test_logout_user(self, client, monkeypatch):
        mocked_thing = mock.Mock()
        monkeypatch.setattr(keycloak_client.keycloak.KeycloakOpenID, "logout", mocked_thing)
        client.logout_user("token")
        mocked_thing.assert_called_once_with("token")

    def test_userinfo(self, client, monkeypatch):
        monkeypatch.setattr(keycloak_client.keycloak.KeycloakOpenID, "userinfo", mock.Mock())

        client.userinfo("token")

        keycloak_client.keycloak.KeycloakOpenID.userinfo.assert_called_once_with("token")

    def test_decode_token(self, client, monkeypatch):
        monkeypatch.setattr(keycloak_client.keycloak.KeycloakOpenID, "decode_token", mock.Mock())

        client.decode_token("token", "key")

        keycloak_client.keycloak.KeycloakOpenID.decode_token.assert_called_once_with("token", "key")

    @pytest.mark.parametrize(
        ("error", "expected"),
        [
            ('{"description": "thing"}', {"description": "thing"}),
            ("{unparsable JSON}", "{unparsable JSON}"),
            ("just some string", "just some string"),
        ],
    )
    def test_parse_error(self, client, error, expected):
        assert client.parse_error(keycloak.KeycloakError(error)) == expected



import pathlib
from dataclasses import dataclass
from unittest import mock

from neos_common.client import email

template_path = pathlib.Path(__file__).resolve().parent / "templates"


@dataclass
class EmailVerify(email.TransactionalEmail):
    template_path = template_path
    template_name = "email_verify"
    subject = "Verify Email."
    token: str
    hostname: str


class TestTransactionalEmail:
    def test_message_generation(self):
        t = EmailVerify(
            to_email="to-email",
            to_name="to-name",
            from_name="from-name",
            from_email="from-email",
            token="ABC123",
            hostname="localhost",
        )
        assert t.message == "Please go to localhost and enter code ABC123."


class TestSMTPEmailHandler:
    def test_init(self):
        eh = email.SMTPEmailHandler(
            config=email.SMTPConfig(
                smtp_server="smtp.server.com",
                smtp_port=25,
                smtp_username="username",
                smtp_password="password",
            ),
            from_email="from_email",
            from_name="from_name",
        )

        assert eh.smtp_server == "smtp.server.com"
        assert eh.smtp_port == 25
        assert eh.smtp_username == "username"
        assert eh.smtp_password == "password"
        assert eh.from_email == "from_email"
        assert eh.from_name == "from_name"

    def test_client_property(self, monkeypatch):
        client = mock.Mock()
        monkeypatch.setattr(
            email.smtplib,
            "SMTP",
            mock.Mock(return_value=client),
        )

        eh = email.SMTPEmailHandler(
            config=email.SMTPConfig(
                smtp_server="smtp.server.com",
                smtp_port=25,
                smtp_username="username",
                smtp_password="password",
            ),
            from_email="from_email",
            from_name="from_name",
        )

        c = eh.client
        assert c == client

        assert email.smtplib.SMTP.call_args == mock.call("smtp.server.com", 25)

    def test_send(self, monkeypatch):
        client = mock.Mock()
        monkeypatch.setattr(
            email.smtplib,
            "SMTP",
            mock.Mock(return_value=client),
        )

        eh = email.SMTPEmailHandler(
            config=email.SMTPConfig(
                smtp_server="smtp.server.com",
                smtp_port=25,
                smtp_username="username",
                smtp_password="password",
            ),
            from_email="from_email",
            from_name="from_name",
        )

        eh.setup()

        eh.send(
            eh["EmailVerify"],
            to_email="test@test.com",
            to_name="mike",
            token="ABC123",
            hostname="localhost",
        )

        assert client.sendmail.call_args == mock.call(
            from_addr="from_email",
            to_addrs=["test@test.com"],
            msg="""Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
From: from_email
To: test@test.com
Subject: Verify Email.

Please go to localhost and enter code ABC123.""",
        )



class TokenData:
    """Represent token data contained within a JWT."""

    def __init__(self, user_id: str, auth_token: str, resources: list[str]) -> None:
        """TokenData initiator..

        Args:
        ----
        user_id: User id.
        auth_token: Raw auth token
        resources: Optional resource urns applicable to current request.
        """
        self.user_id = user_id
        self.auth_token = auth_token
        self.resources = resources

    def get_principal(self) -> str:
        """Get principal ID from the token, usually it is the user ID."""
        return self.user_id



import pytest

from neos_common import error
from neos_common.base import ResourceBase, ResourceReader


class Resource(ResourceBase):
    @classmethod
    def get_resource_id_template(cls):
        return "{resource_id}"

    @classmethod
    def format_resource_id(cls, *args):
        return cls.get_resource_id_template().format(resource_id=args[0])


class TestResource:
    def test_resource_creation(self):
        resource = Resource(partition="ksa", service="iam")
        assert resource.urn == "urn:ksa:iam:::"

    def test_resource_creation_with_all_fields(self):
        resource = Resource(
            partition="ksa",
            service="iam",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
        )
        assert resource.urn == "urn:ksa:iam:id:account:resource_type"

    def test_resource_creation_with_resource(self):
        resource = Resource(
            partition="ksa",
            service="iam",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
            resource_id="resource_id",
        )
        assert resource.urn == "urn:ksa:iam:id:account:resource_type:resource_id"

    def test_urn_template(self):
        resource = Resource()
        assert resource.urn_template == "urn:::::"

    def test_urn_template_star(self):
        resource = Resource(all_="*")
        assert resource.urn_template == "*"

    def test_urn_template_with_fields(self):
        resource = Resource(
            partition="ksa",
            service="iam",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
        )
        assert resource.urn_template == "urn:ksa:iam:id:account:resource_type"

    def test_urn_template_with_resource_id(self):
        resource = Resource(
            partition="ksa",
            service="iam",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
            resource_id="resource_id",
        )
        assert resource.urn_template == "urn:ksa:iam:id:account:resource_type:{resource_id}"

    def test_urn_template_with_resource_id_and_sub_type(self):
        resource = Resource(
            partition="ksa",
            service="iam",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
            sub_type="user",
            resource_id="resource_id",
        )
        assert resource.urn_template == "urn:ksa:iam:id:account:resource_type:user:{resource_id}"

    @pytest.mark.parametrize(
        ("resource_id", "expected_resource_id"),
        [
            ("", ""),
            ("string", "string"),
            (("This", "is", "tuple"), "This"),  # See the realization of the Resource
        ],
    )
    async def test_resource_generation(self, resource_id, expected_resource_id):
        resource = Resource.generate(
            partition="ksa",
            service="core",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
            resource_id=resource_id,
        )
        assert resource == Resource(
            partition="ksa",
            service="core",
            identifier="id",
            account_id="account",
            resource_type="resource_type",
            resource_id=expected_resource_id,
        )


class TestResourceParse:
    @pytest.mark.parametrize(
        "resource_str",
        [
            # Allow all
            ("*"),
            # All fields used and urn/nrn
            ("nrn:ksa:core:sc:root:product:my-dp"),
            ("urn:ksa:core:sc:root:product:my-dp"),
            # Allow any resource_id
            ("urn:ksa:core:sc:root:product:*"),
            # Null sector (identifier)
            ("urn:ksa:iam::root:user:ann"),
            # Null owner (account_id)
            ("urn:ksa:iam:sc::user:ann"),
            # Resource_type and resource_id subgroups
            ("urn:ksa:iam::root:policy:user:something"),
            # Resource_type and resource_id all
            ("urn:ksa:iam::root:policy:user:*"),
            # All nulls
            ("urn:ksa:amazingsvc:::amazingrt:amazinrid"),
        ],
    )
    async def test_should_parse(self, resource_str):
        resource = Resource.parse(resource_str)
        assert resource

        assert resource.urn == resource_str

    @pytest.mark.parametrize(
        "resource_str",
        [
            (""),
            ("*rn:ksa:core:sc:root:product:my-dp"),
            ("arn:ksa:core:sc:root:product:my-dp"),
            ("urn:ksa:core:sc:root:product:my-dp-with-*"),
            ("really_bad_resource"),
            # If we ever decide to use pathlike resource_type/resource_id
            ("urn:ksa:core:sc:root:product/*"),
            ("urn:ksa:iam::root:groups:group-id/something"),
            ("urn:ksa:iam::root:groups/group-id/something"),
        ],
    )
    async def test_should_not_parse(self, resource_str):
        with pytest.raises(error.InvalidResourceFormatError):
            Resource.parse(resource_str)


class TestResourceReader:
    @pytest.mark.parametrize(
        ("identifier", "account_id", "resource_id"),
        [
            ("0123456789", "root", "*"),
            ("0123456789", "root", "dp"),
            ("0123456789", "", "dp"),
            ("", "root", "dp"),
            ("0123456789", "root", ""),
            ("0123456789", "", ""),
            ("", "root", ""),
            ("", "", ""),
        ],
    )
    def test_slice(self, identifier, account_id, resource_id):
        urn = f"urn:ksa:core:{identifier}:{account_id}:product:{resource_id}".rstrip(":")
        r = ResourceReader.parse(urn)

        assert len(r) == 7

        assert r[0] == "urn"
        assert r[1] == "ksa"
        assert r[2] == "core"
        assert r[3] == identifier
        assert r[4] == account_id
        assert r[5] == "product"

        assert r[6] == resource_id
        assert r[-1] == resource_id

        assert r[0:-1] == f"urn:ksa:core:{identifier}:{account_id}:product"

        assert r.urn == urn

    @pytest.mark.parametrize(
        ("identifier", "account_id", "resource_id"),
        [
            ("0123456789", "root", "*"),
            ("0123456789", "root", "dp"),
            ("0123456789", "", "dp"),
            ("", "root", "dp"),
        ],
    )
    def test_slice_with_subtype(self, identifier, account_id, resource_id):
        urn = f"urn:ksa:core:{identifier}:{account_id}:product:sub-type:{resource_id}".rstrip(":")
        r = ResourceReader.parse(urn)

        assert len(r) == 8

        assert r[0] == "urn"
        assert r[1] == "ksa"
        assert r[2] == "core"
        assert r[3] == identifier
        assert r[4] == account_id
        assert r[5] == "product"
        assert r[6] == "sub-type"

        assert r[7] == resource_id
        assert r[-1] == resource_id

        assert r[0:-1] == f"urn:ksa:core:{identifier}:{account_id}:product"

        assert r.urn == urn

    def test_slice_star(self):
        urn = "*"
        r = ResourceReader.parse(urn)

        assert len(r) == 1
        assert r[0] == "*"
        assert r.urn == urn
        assert r[0:-1] == ""

    def test_is_anies(self):
        r = ResourceReader.parse("*")

        assert r.is_any()
        assert not r.is_any_resource_id()

        r = ResourceReader.parse("urn:ksa:core:123:root:product:*")

        assert not r.is_any()
        assert r.is_any_resource_id()

    @pytest.mark.parametrize(
        ("resource_str", "expected_root"),
        [
            # Allow all
            ("*", "*"),
            # All fields used and urn/nrn
            ("nrn:ksa:core:sc:account:product:my-dp", "nrn:ksa:core:sc:root:product:my-dp"),
            ("urn:ksa:core:sc:account:product:my-dp", "urn:ksa:core:sc:root:product:my-dp"),
            # Allow any resource_id
            ("urn:ksa:core:sc:account:product:*", "urn:ksa:core:sc:root:product:*"),
            # Null sector (identifier)
            ("urn:ksa:iam::account:user:ann", "urn:ksa:iam::root:user:ann"),
            # Null owner (account_id)
            ("urn:ksa:iam:sc::user:ann", "urn:ksa:iam:sc:root:user:ann"),
            # Resource_type and resource_id subgroups
            ("urn:ksa:iam::account:policy:user:something", "urn:ksa:iam::root:policy:user:something"),
            # All nulls
            ("urn:ksa:amazingsvc:::amazingrt:amazinrid", "urn:ksa:amazingsvc::root:amazingrt:amazinrid"),
        ],
    )
    async def test_should_parse(self, resource_str, expected_root):
        resource = ResourceReader.parse(resource_str)

        root_resource = ResourceReader.to_root(resource)

        assert root_resource.urn == expected_root



from neos_common.error import UnauthorisedError


class TestError(UnauthorisedError):
    message_ = "Test error."
    reason_ = "test-error"


class NoCodeError(UnauthorisedError):
    title = "Test error."

    @property
    def type(self):
        return "no-code"



import logging
from uuid import UUID

import httpx
from aws4.key_pair import KeyPair

from neos_common import error
from neos_common.base import Action
from neos_common.client.base import NeosClient
from neos_common.schema import Contract

logger = logging.getLogger(__name__)


class HubClient(NeosClient):
    """Identity Access Manager and Registry client."""

    @property
    def service_name(self) -> str:
        return "Hub"

    @property
    def known_errors(self) -> set[str]:
        return {
            "policy-not-found",
            "core-already-registered",
            "core-not-registered",
            "data-product-already-registered",
            "data-product-not-registered",
            "core-already-subscribed",
            "core-not-subscribed",
            "core-host-not-defined",
            "data-product-has-active-subscriptions",
            "core-service-user-not-defined",
            "user-manager-service-error",
        }

    def process_response_with_mapping(self, response: httpx.Response) -> dict:
        try:
            return self.process_response(response)
        except (error.ServiceApiError, error.UnhandledServiceApiError) as exc:
            if exc.reason == "authorization-required":
                raise error.AuthorizationRequiredError(exc.debug) from exc
            if exc.reason == "invalid-authorization":
                raise error.InvalidAuthorizationError(exc.debug) from exc
            if exc.reason == "insufficient-permissions":
                raise error.InsufficientPermissionsError(exc.debug) from exc
            if exc.reason == "invalid-resource-format":
                raise error.InvalidResourceFormatError(exc.debug) from exc
            if exc.reason == "identity-access-manager-error":
                raise error.IdentityAccessManagerError(exc.debug) from exc
            if exc.reason == "signature-error":
                raise error.SignatureError(exc.debug) from exc
            raise

    def __init__(
        self,
        host: str,
        token: str | None,
        key_pair: KeyPair | None,
        account: str,
        partition: str,
    ) -> None:
        # TODO(edgy): Use a better error
        assert token is not None or key_pair is not None  # noqa: S101

        self._token = token
        self._key_pair = key_pair

        self._account = account
        self._partition = partition

        self._host = host
        self._principals = None

    @property
    def token(self) -> str | None:
        return self._token

    @property
    def key_pair(self) -> KeyPair | None:
        return self._key_pair

    @property
    def partition(self) -> str:
        return self._partition

    @property
    def iam_host(self) -> str:
        return f"{self._host}/iam"

    @property
    def registry_host(self) -> str:
        return f"{self._host}/registry"

    @property
    def validate_token_url(self) -> str:
        return f"{self.iam_host}/validate/token"

    @property
    def validate_signature_url(self) -> str:
        return f"{self.iam_host}/validate/signature"

    @property
    def core_url(self) -> str:
        return f"{self.registry_host}/core/{{identifier}}/announce"

    @property
    def data_product_url(self) -> str:
        return f"{self.registry_host}/core/{{identifier}}/data_product"

    @property
    def metadata_url(self) -> str:
        return f"{self.registry_host}/core/{{identifier}}/data_product/metadata"

    async def validate_token(
        self,
        principal: UUID,
        actions: list[Action],
        resources: list[str],
        logic_operator: str,
        *,
        return_allowed_resources: bool = False,
    ) -> tuple[UUID, list[str]]:
        r = await self._get(
            url=self.validate_token_url,
            params={
                "principal_id": principal,
                "action": [action.value for action in actions],
                "resource": resources,
                "logic_operator": logic_operator,
                "return_allowed_resources": return_allowed_resources,
            },
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        data = self.process_response_with_mapping(r)
        logger.info(data)
        return data["principal_id"], data["resources"]

    async def validate_signature(
        self,
        access_key: str,
        auth_schema: str,
        scope: str,
        challenge: str,
        signed_challenge: str,
        actions: list[Action],
        resources: list[str],
        logic_operator: str,
        *,
        return_allowed_resources: bool = False,
    ) -> tuple[UUID, list[str]]:
        r = await self._get(
            url=self.validate_signature_url,
            params={
                "access_key_id": access_key,
                "auth_schema": auth_schema,
                "scope": scope,
                "challenge": challenge,
                "signed_challenge": signed_challenge,
                "action": [action.value for action in actions],
                "resource": resources,
                "logic_operator": logic_operator,
                "return_allowed_resources": return_allowed_resources,
            },
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        data = self.process_response_with_mapping(r)

        return data["principal_id"], data["resources"]

    async def get_allowed_resources(
        self,
        principal_id: str,
        action: Action,
    ) -> list[str]:
        r = await self._get(
            f"{self.iam_host}/principal/{principal_id}/resource",
            params={"action": action.value},
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        d = self.process_response(r)

        return d["resources"]

    async def announce_core(self, host: str, version: str, core_identifier: str) -> None:
        r = await self._post(
            url=self.core_url.format(identifier=core_identifier),
            json={
                "host": host,
                "version": version,
            },
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        self.process_response(r)

    async def register_data_product(
        self,
        urn: str,
        name: str,
        metadata: dict,
        description: str,
        label: str,
        sanitized_name: str,
        data_product_type: str,
        engine: str,
        table: str,
        identifier: str,
        core_identifier: str,
        *,
        contract: Contract,
    ) -> None:
        """Registers a data product by sending a request to the specified endpoint.

        Args:
            urn (str): Unique resource name.
            name (str): Data product name.
            metadata (dict): Additional metadata.
            description (str): Brief data product description.
            label (str): Product label or category.
            sanitized_name (str): Sanitized version of the name.
            data_product_type (str): Type of data product.
            engine (str): Data engine used.
            table (str): Table or collection name.
            identifier (str): Unique data product identifier.
            core_identifier (str): Identifier for URL formatting.
            contract (Contract): Contract defining visibility and subscription.

        Returns:
            None: Function returns None.

        """
        json = {
            "urn": urn,
            "name": name,
            "metadata": metadata,
            "description": description,
            "label": label,
            "sanitized_name": sanitized_name,
            "data_product_type": data_product_type,
            "engine": engine,
            "table": table,
            "identifier": identifier,
            "contract": contract.model_dump(mode="json"),
        }
        logger.info(
            f"Register data product request: {json}",
        )

        r = await self._post(
            url=self.data_product_url.format(identifier=core_identifier),
            json=json,
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        self.process_response(r)

    async def update_data_product_metadata(
        self,
        urn: str,
        metadata: dict,
        description: str,
        core_identifier: str,
    ) -> None:
        r = await self._post(
            url=self.metadata_url.format(identifier=core_identifier),
            json={
                "urn": urn,
                "metadata": metadata,
                "description": description,
            },
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        self.process_response(r)

    async def deregister_data_product(self, urn: str, core_identifier: str) -> None:
        r = await self._delete(
            url=self.data_product_url.format(identifier=core_identifier),
            json={
                "urn": urn,
            },
            headers={"X-Account": self._account, "X-Partition": self._partition},
        )

        self.process_response(r)



"""Collection of useful commands for `neosctl` management.

To view a list of available commands:

$ invoke --list
"""

import invoke


@invoke.task
def install(context):
    """Install production requirements for `neos_common`."""
    context.run("uv sync --extra kafka --extra config --extra auth")


@invoke.task
def install_dev(context):
    """Install development requirements for `neos_common`."""
    context.run("uv sync --all-extras")
    context.run("uv run pre-commit install")
    context.run(
        """
        echo "Generating pyrightconfig.json";
        echo "{\\"venv\\": \\".\\", \\"venvPath\\": \\".venv)\\", \\"exclude\\": [\\"tests\\"], \\"include\\": [\\"neos_common\\"]}" > pyrightconfig.json
    """,
    )


@invoke.task
def check_style(context):
    """Run style checks."""
    context.run("ruff .")


@invoke.task
def check_types(context):
    """Run pyright checks."""
    context.run("pyright neos_common")


@invoke.task
def tests(context):
    """Run pytest unit tests."""
    context.run("pytest -x -s")


@invoke.task
def tests_debug(context):
    """Run pytest unit tests with debug logs."""
    context.run("pytest -x -s -o log_cli=1 -o log-cli-level=DEBUG")


@invoke.task
def tests_coverage(context, output="term-missing"):
    """Run pytest unit tests with coverage.

    Coverage when plugins are involved gets funky, without this coverage is reported at 50-60% instead of 100%
    https://pytest-cov.readthedocs.io/en/latest/plugins.html
    """
    context.run(
        f"pytest --cov=neos_common -x --cov-report={output}",
    )


@invoke.task
def release(context):
    """Bump to next X.Y.Z version."""
    context.run("changelog generate")


@invoke.task
def bump_patch(context):
    """Bump to next X.Y.patch version."""
    context.run("changelog generate --version-part=patch")


@invoke.task
def bump_minor(context):
    """Bump to next X.minor.0 version."""
    context.run("changelog generate --version-part=minor")


@invoke.task
def bump_major(context):
    """Bump to next major.0.0 version."""
    context.run("changelog generate --version-part=major")



from unittest import mock

import pytest
from aws4.key_pair import KeyPair

from neos_common import base, config, error
from tests.conftest import AsyncMock

keycloak = pytest.importorskip("keycloak")
util = pytest.importorskip("neos_common.authorization.util")


class Config(config.ApiConfig):
    name: str = "test"

    access_manager_host: str = "https://access-manager-host"


@pytest.fixture
def config_obj():
    return Config(
        keycloak_host="https://keycloak-host",
        keycloak_realm="test",
        keycloak_client_id="client-id",
        keycloak_client_secret="client-secret",
        postgres_host="postgres-host",
        postgres_user="postgres-user",
        postgres_password="postgres-password",
        postgres_database="postgres-database",
        access_key_id="access-key-id",
        secret_access_key="secret-access-key",
    )


class TestCommonConfig:
    def test_api_prefix_prepends_slash_if_missing(self, config_obj, monkeypatch):
        monkeypatch.setattr(config_obj, "raw_api_prefix", "api")

        assert config_obj.api_prefix == "/api"

    def test_api_prefix_handles_existing_slash(self, config_obj, monkeypatch):
        monkeypatch.setattr(config_obj, "raw_api_prefix", "/api")

        assert config_obj.api_prefix == "/api"

    def test_logging_config(self, config_obj):
        assert config_obj.logging_config == {
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                    "format": "[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                },
            },
            "handlers": {
                "default": {
                    "class": "logging.StreamHandler",
                    "formatter": "default",
                    "level": "INFO",
                },
            },
            "loggers": {
                "test": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "uvicorn.access": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "pogo_migrate": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "neos_common": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
            },
            "version": 1,
        }


class TestOpenapiInfo:
    def test_info_creation(self, config_obj):
        info_dict: dict[str, str] = {
            "x-iam-action-0": "core:announce",
            "x-iam-resource-0": "test:core:resource_id",
            "logic-operator": "and",
        }

        def generate_from_config(config, resource_type, resource_id):
            return mock.Mock(urn=f"{config.name}:{resource_type.value}:{resource_id}")

        assert info_dict == util.create_openapi_info(
            config_obj,
            mock.Mock(generate_from_config=generate_from_config),
            (base.Action.core_announce, base.ResourceType.core, "resource_id"),
        )

    def test_info_creation_with_account(self, config_obj):
        info_dict: dict[str, str] = {
            "x-iam-action-0": "core:announce",
            "x-iam-resource-0": "test:ksa:account:core:resource_id",
            "logic-operator": "or",
        }

        def generate_from_config(config, partition, account_id, resource_type, resource_id):
            return mock.Mock(urn=f"{config.name}:{partition}:{account_id}:{resource_type.value}:{resource_id}")

        assert info_dict == util.create_openapi_info(
            config_obj,
            mock.Mock(generate_from_config=generate_from_config),
            (base.Action.core_announce, base.ResourceType.core, "resource_id", "ksa", "account"),
            logic_operator="or",
        )

    def test_info_creation_invalid_call(self, config_obj):
        with pytest.raises(ValueError, match="Invalid number of arguments"):
            util.create_openapi_info(
                config_obj,
                mock.Mock(),
                (base.Action.core_announce, base.ResourceType.core, "resource_id", "ksa", "account", "unexpected"),
            )


class TestDepKeycloakClientFactory:
    async def test_client_returned(self, config_obj, monkeypatch):
        kc = mock.Mock()
        monkeypatch.setattr(util, "KeycloakClient", mock.Mock(return_value=kc))

        DepKeycloakClient = util.DepKeycloakClientFactory.build(mock.Mock())

        client = await DepKeycloakClient()(config_obj)

        assert client == kc
        assert util.KeycloakClient.call_args == mock.call(
            host=config_obj.keycloak_host,
            realm=config_obj.keycloak_realm,
            client_id=config_obj.keycloak_client_id,
            client_secret=config_obj.keycloak_client_secret,
        )


class Resource(base.ResourceBase):
    def generate_from_config(self, *args, **kwargs) -> str:  # noqa: ARG002
        return ""


class TestDepAuthorizationFactory:
    def test_init_single_action_resource(self):
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization(util.ActionResource("action", "resource_type", "resource"))

        assert dep.action_resources == util.ActionResourceList(
            [util.ActionResource("action", "resource_type", "resource")],
            "and",
        )

    def test_init_action_resources(self):
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization([util.ActionResource("action", "resource_type", "resource")])

        assert dep.action_resources == util.ActionResourceList(
            [util.ActionResource("action", "resource_type", "resource")],
            "and",
        )

    def test_init_action_resource_list(self):
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization(
            util.ActionResourceList([util.ActionResource("action", "resource_type", "resource")], "or"),
        )

        assert dep.action_resources == util.ActionResourceList(
            [util.ActionResource("action", "resource_type", "resource")],
            "or",
        )

    async def test_no_token_enforced(self, config_obj):
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization(util.ActionResourceList([], "or"))
        request = mock.Mock(headers={})

        with pytest.raises(error.AuthorizationRequiredError):
            await dep(request, mock.Mock(), config_obj, mock.Mock(), mock.Mock())

    async def test_keycloak_error_handled(self, config_obj):
        mock_keycloak = mock.Mock(
            introspect_token=mock.Mock(side_effect=keycloak.KeycloakError("an error")),
        )
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization(
            util.ActionResourceList([util.ActionResource("action", base.ResourceType.core, Resource)], "and"),
        )
        request = mock.Mock(headers={"Authorization": "Bearer token"})

        with pytest.raises(error.IdentityAccessManagerError):
            await dep(request, mock_keycloak, config_obj, mock.Mock(), mock.Mock())

    async def test_inactive_token(self, config_obj):
        mock_keycloak = mock.Mock(
            introspect_token=mock.Mock(
                return_value={"active": False},
            ),
        )
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization([])
        request = mock.Mock(headers={"Authorization": "Bearer token"})

        with pytest.raises(error.InvalidAuthorizationError):
            await dep(request, mock_keycloak, config_obj, mock.Mock(), mock.Mock())

    async def test_malformed_token(self, config_obj):
        mock_keycloak = mock.Mock(
            introspect_token=mock.Mock(
                return_value={"active": True},
            ),
        )
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization([])
        request = mock.Mock(headers={"Authorization": "Bearer token"})

        with pytest.raises(error.InvalidAuthorizationError) as e:
            await dep(request, mock_keycloak, config_obj, mock.Mock(), mock.Mock())

        assert e.value.debug == "Invalid token format, 'sub'"

    async def test_valid_token(self, config_obj):
        mock_keycloak = mock.Mock(
            introspect_token=mock.Mock(
                return_value={
                    "active": True,
                    "sub": "user-id",
                },
            ),
        )
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization([])
        request = mock.Mock(headers={"Authorization": "Bearer token"})

        token = await dep(request, mock_keycloak, config_obj, mock.Mock(), mock.Mock())

        assert token.user_id == "user-id"

    async def test_valid_signature(self, config_obj):
        mock_keycloak = mock.Mock()
        mock_signature = mock.Mock(validate=AsyncMock(return_value=("user-id", [])))
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization([])
        request = mock.Mock(headers={"Authorization": "NEOS4-HMAC-SHA256 ..."})

        token = await dep(request, mock_keycloak, config_obj, mock_signature, mock.Mock())

        assert token.user_id == "user-id"
        assert token.resources == []

    async def test_unsupported_auth(self, config_obj):
        mock_keycloak = mock.Mock()
        DepAuthorization = util.DepAuthorizationFactory.build(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        dep = DepAuthorization([])
        request = mock.Mock(headers={"Authorization": "Basic token"})

        with pytest.raises(error.InvalidAuthorizationError):
            await dep(request, mock_keycloak, config_obj, mock.Mock(), mock.Mock())


class TestDepAccessValidatorFactory:
    async def test_validator_returned(self, monkeypatch):
        hub_client = mock.Mock()
        av = mock.Mock()
        monkeypatch.setattr(util.validator, "AccessValidator", mock.Mock(return_value=av))
        DepAccessValidator = util.DepAccessValidatorFactory.build(mock.Mock())
        access_validator = await DepAccessValidator()(hub_client)

        assert access_validator == av

    async def test_default_client_generated(self):
        hub_client = mock.Mock()
        DepAccessValidator = util.DepAccessValidatorFactory.build(mock.Mock())
        access_validator = await DepAccessValidator()(hub_client)

        assert access_validator._hub_client == hub_client

    async def test_validator_generates_client(self):
        hub_client = mock.Mock()

        DepAccessValidator = util.DepAccessValidatorFactory.build(mock.Mock())
        access_validator = await DepAccessValidator()(hub_client)

        assert access_validator._hub_client == hub_client


class TestDepHubClientFactory:
    async def test_hub_client_returned(self, monkeypatch):
        config = mock.Mock(
            hub_host="host",
            access_key_id="access-key",
            secret_access_key="secret-key",
            partition="ksa",
            account="root",
        )
        hub_client_mock = mock.Mock()
        HubClientMock = mock.Mock(return_value=hub_client_mock)
        monkeypatch.setattr(util, "HubClient", HubClientMock)
        DepHubClient = util.DepHubClientFactory.build(mock.Mock())
        hub_client = await DepHubClient()(config)

        assert HubClientMock.call_args == mock.call(
            host="host",
            token=None,
            key_pair=KeyPair(
                access_key_id="access-key",
                secret_access_key="secret-key",
            ),
            account="root",
            partition="ksa",
        )
        assert hub_client == hub_client_mock



import importlib.metadata
from unittest import mock

import pydantic
import pytest

from neos_common import error
from neos_common.socket import client
from tests.conftest import AsyncMock

RAW_MESSAGE = b'{"hello": "world"}'
MESSAGE_HEADER = b"\x00\x00\x00\x12"

PYDANTIC_VERSION = ".".join(importlib.metadata.version("pydantic").split(".")[:2])


class RequestData(pydantic.BaseModel):
    name: str


class TestTCPClient:
    def test_init(self):
        c = client.TCPClient("host", "port", "timeout", "wait")

        assert c.addr == ("host", "port")
        assert c.timeout == "timeout"
        assert c.wait == "wait"

    def test_send_request_creates_connection(self, monkeypatch):
        sock = mock.Mock()
        monkeypatch.setattr(client.socket, "create_connection", mock.Mock(return_value=sock))
        monkeypatch.setattr(sock, "recv", mock.Mock(side_effect=[MESSAGE_HEADER, RAW_MESSAGE]))

        c = client.TCPClient("host", "port", 10, 1)

        c.send_request("request")

        assert client.socket.create_connection.call_args == mock.call(("host", "port"), timeout=10)

    def test_send_request_handles_connection_error(self, monkeypatch):
        monkeypatch.setattr(
            client.socket,
            "create_connection",
            mock.Mock(side_effect=client.socket.gaierror("[Err 2]: fail")),
        )

        c = client.TCPClient("host", "port", 10, 1)

        with pytest.raises(error.ServiceConnectionError) as e:
            c.send_request("request")

        assert e.value.debug == "('host', 'port'): [Err 2]: fail"
        assert e.value.message == "Unable to connect to service."

    def test_send_request_sends_message(self, monkeypatch):
        sock = mock.Mock()
        monkeypatch.setattr(client.socket, "create_connection", mock.Mock(return_value=sock))
        monkeypatch.setattr(sock, "recv", mock.Mock(side_effect=[MESSAGE_HEADER, RAW_MESSAGE]))

        c = client.TCPClient("host", "port", 10, 1)

        c.send_request("request")

        assert sock.sendall.call_args == mock.call(b'\x00\x00\x00\t"request"')

    def test_send_request_returns_response(self, monkeypatch):
        sock = mock.Mock()
        monkeypatch.setattr(client.socket, "create_connection", mock.Mock(return_value=sock))
        monkeypatch.setattr(sock, "recv", mock.Mock(side_effect=[MESSAGE_HEADER, RAW_MESSAGE]))

        c = client.TCPClient("host", "port", 10, 1)

        r = c.send_request("request")

        assert r == {"hello": "world"}

    def test_send_request_timeout(self, monkeypatch):
        sock = mock.Mock()
        monkeypatch.setattr(client.socket, "create_connection", mock.Mock(return_value=sock))
        monkeypatch.setattr(sock, "recv", mock.Mock(return_value=None))

        c = client.TCPClient("host", "port", 10, 0)

        with pytest.raises(TimeoutError):
            c.send_request("request")


class TestAsyncTCPClient:
    def test_init(self):
        c = client.AsyncTCPClient("host", "port", "encoder")

        assert c.addr == ("host", "port")
        assert c.encoder == "encoder"

    async def test_send_request_connection_error(self, monkeypatch):
        monkeypatch.setattr(client.asyncio, "open_connection", AsyncMock(side_effect=OSError("connection issue")))
        c = client.AsyncTCPClient("host", "port")

        with pytest.raises(error.ServiceConnectionError) as e:
            await c.send_request({"hello": "world"})

        assert e.value.debug == "('host', 'port'): connection issue"
        assert e.value.message == "Unable to connect to service."

        assert client.asyncio.open_connection.call_args == mock.call("host", "port")

    async def test_request_sent(self, monkeypatch):
        reader, writer = AsyncMock(), mock.Mock()
        monkeypatch.setattr(client.asyncio, "open_connection", AsyncMock(return_value=(reader, writer)))
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"world": "hello"}',
        ]

        c = client.AsyncTCPClient("host", "port")

        await c.send_request({"hello": "world"})

        assert writer.write.call_args == mock.call(
            b'\x00\x00\x00\x12{"hello": "world"}',
        )

    async def test_response_received(self, monkeypatch):
        reader, writer = AsyncMock(), mock.Mock()
        monkeypatch.setattr(client.asyncio, "open_connection", AsyncMock(return_value=(reader, writer)))
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"world": "hello"}',
        ]

        c = client.AsyncTCPClient("host", "port")

        r = await c.send_request({"hello": "world"})

        assert r == {"world": "hello"}

    async def test_response_received_multipart(self, monkeypatch):
        reader, writer = AsyncMock(), mock.Mock()
        monkeypatch.setattr(client.asyncio, "open_connection", AsyncMock(return_value=(reader, writer)))
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"worl',  # pragma: no-spell-check
            b'd": "hello"}',
        ]

        c = client.AsyncTCPClient("host", "port")

        r = await c.send_request({"hello": "world"})

        assert r == {"world": "hello"}



import importlib.metadata
from unittest import mock

import pydantic
import pytest

from neos_common.socket import server, util
from tests.conftest import AsyncMock

RAW_MESSAGE = b'{"hello": "world"}'
MESSAGE_HEADER = b"\x00\x00\x00\x12"

PYDANTIC_VERSION = ".".join(importlib.metadata.version("pydantic").split(".")[:2])


class RequestData(pydantic.BaseModel):
    name: str


class DummyTCPHandler(server.TCPHandler):
    def setup(self):
        self.request_mapping = {
            "test": RequestData,
        }


class TestTCPHandler:
    def test_handle_invalid_request(self):
        request = mock.Mock()
        request.recv.side_effect = [
            MESSAGE_HEADER,
            b'{"hello": "world"}',
        ]

        DummyTCPHandler(request, None, None)

        assert request.sendall.call_args == mock.call(
            b"\x00\x00\x01\xb7"
            + bytes(
                f'{{"ok": false, "status": "unprocessible-entity", "reason": "ValidationError", "message": "Unable to parse message.", "debug": null, "details": [{{"type": "missing", "loc": ["request_type"], "msg": "Field required", "input": {{"hello": "world"}}, "url": "https://errors.pydantic.dev/{PYDANTIC_VERSION}/v/missing"}}, {{"type": "missing", "loc": ["data"], "msg": "Field required", "input": {{"hello": "world"}}, "url": "https://errors.pydantic.dev/{PYDANTIC_VERSION}/v/missing"}}]}}',
                "utf-8",
            ),
        )

    def test_handle_invalid_request_data(self):
        request = mock.Mock()
        request.recv.side_effect = [
            MESSAGE_HEADER,
            b'{"request_type": "test", "data": {}}',
        ]

        DummyTCPHandler(request, None, None)

        assert request.sendall.call_args == mock.call(
            b"\x00\x00\x01\x0f"
            + bytes(
                f'{{"ok": false, "status": "unprocessible-entity", "reason": "ValidationError", "message": "Unable to parse message.", "debug": null, "details": [{{"type": "missing", "loc": ["name"], "msg": "Field required", "input": {{}}, "url": "https://errors.pydantic.dev/{PYDANTIC_VERSION}/v/missing"}}]}}',
                "utf-8",
            ),
        )

    def test_handle_receive_exception(self):
        request = mock.Mock()
        request.recv.side_effect = Exception

        DummyTCPHandler(request, None, None)

        assert request.sendall.call_args == mock.call(
            b'\x00\x00\x00j{"ok": false, "status": "unhandled", "reason": "Unhandled", "message": "", "debug": null, "details": null}',
        )

    def test_handle_process_exception(self, monkeypatch):
        request = mock.Mock()
        request.recv.side_effect = [
            MESSAGE_HEADER,
            b'{"request_type": "test", "data": {}}',
        ]
        monkeypatch.setattr(DummyTCPHandler, "request_mapping", None)
        monkeypatch.setattr(DummyTCPHandler, "setup", mock.Mock())

        DummyTCPHandler(request, None, None)

        assert request.sendall.call_args == mock.call(
            b'\x00\x00\x00\x92{"ok": false, "status": "unhandled", "reason": "Unhandled", "message": "\'NoneType\' object has no attribute \'get\'", "debug": null, "details": null}',
        )

    def test_request_handled(self, monkeypatch):
        request = mock.Mock()
        request.recv.side_effect = [
            MESSAGE_HEADER,
            b'{"request_type": "test", "data": {"name": "testdata"}}',
        ]

        mock_handle = mock.Mock()
        monkeypatch.setattr(DummyTCPHandler, "_handle", mock_handle)

        DummyTCPHandler(request, None, None)

        assert mock_handle.call_args == mock.call(request, "test", RequestData(name="testdata"))


@pytest.fixture
def handler():
    return server.AsyncTCPHandler()


class TestAsyncHandler:
    async def test_non_json_read(self, handler):
        reader, writer = AsyncMock(), mock.Mock()
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'("hello", "world")',
        ]

        await handler(reader, writer)

        assert writer.write.call_args == mock.call(
            b'\x00\x00\x00\xa0{"ok": false, "status": "unprocessible-entity", "reason": "DecodeError", "message": "Expecting value: line 1 column 1 (char 0)", "debug": null, "details": null}',
        )

    async def test_exception_read(self, handler):
        reader, writer = AsyncMock(), mock.Mock()
        reader.read.side_effect = Exception("exception")

        await handler(reader, writer)

        assert writer.write.call_args == mock.call(
            b'\x00\x00\x00s{"ok": false, "status": "unhandled", "reason": "Unhandled", "message": "exception", "debug": null, "details": null}',
        )

    async def test_data_validation_error(self, handler):
        reader, writer = AsyncMock(), mock.Mock()
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"hello": "world"}',
        ]

        await handler(reader, writer)

        assert writer.write.call_args == mock.call(
            b"\x00\x00\x01\xb7"
            + bytes(
                f'{{"ok": false, "status": "unprocessible-entity", "reason": "ValidationError", "message": "Unable to parse message.", "debug": null, "details": [{{"type": "missing", "loc": ["request_type"], "msg": "Field required", "input": {{"hello": "world"}}, "url": "https://errors.pydantic.dev/{PYDANTIC_VERSION}/v/missing"}}, {{"type": "missing", "loc": ["data"], "msg": "Field required", "input": {{"hello": "world"}}, "url": "https://errors.pydantic.dev/{PYDANTIC_VERSION}/v/missing"}}]}}',
                "utf-8",
            ),
        )

    async def test_processing_exception(self, handler):
        reader, writer = AsyncMock(), mock.Mock()
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"request_type": "test", "data": {"hello": "world"}}',
        ]
        handler.request_mapping = {"test": list}

        await handler(reader, writer)

        assert writer.write.call_args == mock.call(
            b'\x00\x00\x00\x8b{"ok": false, "status": "unhandled", "reason": "Unhandled", "message": "list() takes no keyword arguments", "debug": null, "details": null}',
        )

    async def test_valid_request_handled(self, handler, monkeypatch):
        reader, writer = AsyncMock(), mock.Mock()
        asocket = util.AsyncSocket(reader, writer)
        monkeypatch.setattr(util, "AsyncSocket", mock.Mock(return_value=asocket))
        reader.read.side_effect = [
            MESSAGE_HEADER,
            b'{"request_type": "test", "data": {"hello": "world"}}',
        ]
        monkeypatch.setattr(handler, "_handle", AsyncMock(return_value=None))
        handler.request_mapping = {}

        await handler(reader, writer)

        assert handler._handle.call_args == mock.call(asocket, "test", {"hello": "world"})



from unittest import mock

import pytest
from freezegun import freeze_time

from neos_common import base
from tests.authorization.conftest import Resource
from tests.conftest import AsyncMock

validator = pytest.importorskip("neos_common.authorization.validator")


class TestAccessValidator:
    async def test_init(self):
        access_validator = validator.AccessValidator("hub_client")
        assert access_validator._hub_client == "hub_client"

    async def test_validate(self):
        hub_client = mock.Mock(validate_token=AsyncMock(return_value=("1", ["*"])))

        access_validator = validator.AccessValidator(hub_client)

        action = base.Action("core:announce")
        resource = Resource.generate(
            partition="test",
            service="iam",
            identifier="id",
            account_id="root",
            resource_type="policy",
        )
        result = await access_validator.validate(
            "1",
            [action],
            [resource],
            logic_operator="and",
            return_allowed_resources=True,
        )
        assert result == ("1", ["*"])
        hub_client.validate_token.assert_called_once_with(
            principal="1",
            actions=[action],
            resources=[resource.urn],
            logic_operator="and",
            return_allowed_resources=True,
        )


class TestSignatureValidator:
    @freeze_time("2023-01-01 12:00:00")
    async def test_validate(self):
        mock_hub = AsyncMock()
        action = base.Action("core:announce")
        resource = Resource.generate(
            partition="test",
            service="iam",
            identifier="id",
            account_id="root",
            resource_type="policy",
        )

        request = mock.Mock(
            method="GET",
            url=mock.Mock(scheme="http", path="/", query=b"foo=bar"),
            headers={
                "x-amz-date": "20230101T120000Z",
                "x-amz-content-sha256": "f4eb19f40510b16354e25f8b339dca7a40e44dfb846214371c054677c42146d7",
                "Authorization": "AWS4-HMAC-SHA256 Credential=access-key/20230101/ksa/iam/aws4_request, SignedHeaders=x-amz-date, Signature=signature",
            },
            body=AsyncMock(return_value=b'{"foo": "bar"}'),
        )

        v = validator.SignatureValidator(mock_hub)

        await v.validate(request, [action], [resource], logic_operator="and")

        assert mock_hub.validate_signature.call_args == mock.call(
            "access-key",
            "AWS4",
            "20230101/ksa/iam/aws4_request",
            "AWS4-HMAC-SHA256\n20230101T120000Z\n20230101/ksa/iam/aws4_request\nf4eb19f40510b16354e25f8b339dca7a40e44dfb846214371c054677c42146d7",
            "signature",
            [action],
            [resource.urn],
            logic_operator="and",
            return_allowed_resources=False,
        )

    @freeze_time("2023-01-01 12:00:00")
    async def test_validate_neos(self):
        mock_hub = AsyncMock()
        action = base.Action("core:announce")
        resource = Resource.generate(
            partition="test",
            service="iam",
            identifier="id",
            account_id="root",
            resource_type="policy",
        )

        request = mock.Mock(
            method="GET",
            url=mock.Mock(scheme="http", path="/", query=b"foo=bar"),
            headers={
                "x-neos-date": "20230101T120000Z",
                "x-neos-content-sha256": "f4eb19f40510b16354e25f8b339dca7a40e44dfb846214371c054677c42146d7",
                "Authorization": "NEOS4-HMAC-SHA256 Credential=access-key/20230101/ksa/iam/neos4_request, SignedHeaders=x-neos-date, Signature=signature",
            },
            body=AsyncMock(return_value=b'{"foo": "bar"}'),
        )

        v = validator.SignatureValidator(mock_hub)

        await v.validate(request, [action], [resource], logic_operator="and")

        assert mock_hub.validate_signature.call_args == mock.call(
            "access-key",
            "NEOS4",
            "20230101/ksa/iam/neos4_request",
            "NEOS4-HMAC-SHA256\n20230101T120000Z\n20230101/ksa/iam/neos4_request\n08080976e88185ead32bc4fa943273bd050be5b639c3ba38115dfb9107c3221d",
            "signature",
            [action],
            [resource.urn],
            logic_operator="and",
            return_allowed_resources=False,
        )

    @freeze_time("2023-01-01 12:00:00")
    async def test_validate_binary_body(self):
        mock_hub = AsyncMock()
        action = base.Action("core:announce")
        resource = Resource.generate(
            partition="test",
            service="iam",
            identifier="id",
            account_id="root",
            resource_type="policy",
        )

        body = b"PAR1\x15\x00\x15\x16\x150,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x011\x18\x011\x16\x00(\x011\x18\x011\x00\x00\x00(\xb5/\xfd\x04\x00Y\x00\x00\x02\x00\x00\x00\x02\x01\x01\x00\x00\x001\xa4%\xf8\xcb\x15\x00\x15\x16\x150,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x012\x18\x012\x16\x00(\x012\x18\x012\x00\x00\x00(\xb5/\xfd\x04\x00Y\x00\x00\x02\x00\x00\x00\x02\x01\x01\x00\x00\x002\xf1\xee\xee1\x15\x00\x15\x16\x150,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x013\x18\x013\x16\x00(\x013\x18\x013\x00\x00\x00(\xb5/\xfd\x04\x00Y\x00\x00\x02\x00\x00\x00\x02\x01\x01\x00\x00\x003Yk\xbf\xfb\x15\x00\x15\x1e\x158,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x05Apple\x18\x05Apple\x16\x00(\x05Apple\x18\x05Apple\x00\x00\x00(\xb5/\xfd\x04\x00y\x00\x00\x02\x00\x00\x00\x02\x01\x05\x00\x00\x00Apple\x8f\x0b\x1b\xcd\x15\x00\x15 \x15:,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x06Orange\x18\x06Orange\x16\x00(\x06Orange\x18\x06Orange\x00\x00\x00(\xb5/\xfd\x04\x00\x81\x00\x00\x02\x00\x00\x00\x02\x01\x06\x00\x00\x00Orangel\xa2\x1cw\x15\x00\x15 \x15:,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x06Banana\x18\x06Banana\x16\x00(\x06Banana\x18\x06Banana\x00\x00\x00(\xb5/\xfd\x04\x00\x81\x00\x00\x02\x00\x00\x00\x02\x01\x06\x00\x00\x00BananaS\xae\x06\xb9\x15\x00\x15\x18\x152,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x0234\x18\x0234\x16\x00(\x0234\x18\x0234\x00\x00\x00(\xb5/\xfd\x04\x00a\x00\x00\x02\x00\x00\x00\x02\x01\x02\x00\x00\x0034\xf7\xa0\x11j\x15\x00\x15\x18\x152,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x0256\x18\x0256\x16\x00(\x0256\x18\x0256\x00\x00\x00(\xb5/\xfd\x04\x00a\x00\x00\x02\x00\x00\x00\x02\x01\x02\x00\x00\x0056\x9e\xc5\r\xa9\x15\x00\x15\x18\x152,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x0267\x18\x0267\x16\x00(\x0267\x18\x0267\x00\x00\x00(\xb5/\xfd\x04\x00a\x00\x00\x02\x00\x00\x00\x02\x01\x02\x00\x00\x0067\xee\xaa\nL\x15\x00\x15F\x15`,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x16\x00(\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x00\x00\x00(\xb5/\xfd\x04\x00\x19\x01\x00\x02\x00\x00\x00\x02\x01\x19\x00\x00\x002024-12-04T12:05:25+01:00?f5@\x15\x00\x15F\x15`,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x16\x00(\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x00\x00\x00(\xb5/\xfd\x04\x00\x19\x01\x00\x02\x00\x00\x00\x02\x01\x19\x00\x00\x002024-12-04T12:05:25+01:00?f5@\x15\x00\x15F\x15`,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x16\x00(\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x00\x00\x00(\xb5/\xfd\x04\x00\x19\x01\x00\x02\x00\x00\x00\x02\x01\x19\x00\x00\x002024-12-04T12:05:25+01:00?f5@\x15\x00\x15\xea\x01\x15\x84\x02,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x16\x00(k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x00\x00\x00(\xb5/\xfd\x04\x00\xa9\x03\x00\x02\x00\x00\x00\x02\x01k\x00\x00\x00/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\xc4\xd8\x84\t\x15\x00\x15\xea\x01\x15\x84\x02,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x16\x00(k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x00\x00\x00(\xb5/\xfd\x04\x00\xa9\x03\x00\x02\x00\x00\x00\x02\x01k\x00\x00\x00/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\xc4\xd8\x84\t\x15\x00\x15\xea\x01\x15\x84\x02,\x15\x02\x15\x00\x15\x06\x15\x06\x1c\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x16\x00(k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x00\x00\x00(\xb5/\xfd\x04\x00\xa9\x03\x00\x02\x00\x00\x00\x02\x01k\x00\x00\x00/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\xc4\xd8\x84\t\x191\x02\x02\x02\x198\x011\x012\x013\x198\x011\x012\x013\x15\x00\x196\x00\x00\x00\x00\x191\x02\x02\x02\x198\x05Apple\x06Orange\x06Banana\x198\x05Apple\x06Orange\x06Banana\x15\x00\x196\x00\x00\x00\x00\x191\x02\x02\x02\x198\x0234\x0256\x0267\x198\x0234\x0256\x0267\x15\x00\x196\x00\x00\x00\x00\x191\x02\x02\x02\x198\x192024-12-04T12:05:25+01:00\x192024-12-04T12:05:25+01:00\x192024-12-04T12:05:25+01:00\x198\x192024-12-04T12:05:25+01:00\x192024-12-04T12:05:25+01:00\x192024-12-04T12:05:25+01:00\x15\x00\x196\x00\x00\x00\x00\x191\x02\x02\x02\x198k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsxk/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsxk/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x198k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsxk/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsxk/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x15\x00\x196\x00\x00\x00\x00\x19<\x16\x08\x15r\x16\x00\x00\x16z\x15r\x16\x02\x00\x16\xec\x01\x15r\x16\x04\x00\x00\x19<\x16\xde\x02\x15\x9a\x01\x16\x00\x00\x16\xf8\x03\x15\xa4\x01\x16\x02\x00\x16\x9c\x05\x15\xa4\x01\x16\x04\x00\x00\x19<\x16\xc0\x06\x15|\x16\x00\x00\x16\xbc\x07\x15|\x16\x02\x00\x16\xb8\x08\x15|\x16\x04\x00\x00\x19<\x16\xb4\t\x15\xe2\x02\x16\x00\x00\x16\x96\x0c\x15\xe2\x02\x16\x02\x00\x16\xf8\x0e\x15\xe2\x02\x16\x04\x00\x00\x19<\x16\xda\x11\x15\x9a\t\x16\x00\x00\x16\xf4\x1a\x15\x9a\t\x16\x02\x00\x16\x8e$\x15\x9a\t\x16\x04\x00\x00\x15\x02\x19l5\x00\x18\x0fparquet_go_root\x15\n\x00\x15\x0c\x15\x00\x15\x02\x18\x02id%\x00\x15\x00\x15\x00\x15\x00\x1c\x1c\x00\x00\x00\x15\x0c\x15\x00\x15\x02\x18\x05fruit%\x00\x15\x00\x15\x00\x15\x00\x1c\x1c\x00\x00\x00\x15\x0c\x15\x00\x15\x02\x18\x05count%\x00\x15\x00\x15\x00\x15\x00\x1c\x1c\x00\x00\x00\x15\x0c\x15\x00\x15\x02\x18\x0e__extracted_at%\x00\x15\x00\x15\x00\x15\x00\x1c\x1c\x00\x00\x00\x15\x0c\x15\x00\x15\x02\x18\n__filepath%\x00\x15\x00\x15\x00\x15\x00\x1c\x1c\x00\x00\x00\x16\x06\x19\x1c\x19\\&\x08\x1c\x15\x0c\x195\x06\x08\x00\x19\x18\x02id\x15\x0c\x16\x06\x16\x88\x02\x16\xd6\x02&\x08<\x18\x013\x18\x011\x16\x00(\x013\x18\x011\x00\x00\x16\xa6<\x152\x16\xa8-\x15:\x00&\xde\x02\x1c\x15\x0c\x195\x06\x08\x00\x19\x18\x05fruit\x15\x0c\x16\x06\x16\x94\x03\x16\xe2\x03&\xde\x02<\x18\x06Orange\x18\x05Apple\x16\x00(\x06Orange\x18\x05Apple\x00\x00\x16\xd8<\x15<\x16\xe2-\x15r\x00&\xc0\x06\x1c\x15\x0c\x195\x06\x08\x00\x19\x18\x05count\x15\x0c\x16\x06\x16\xa6\x02\x16\xf4\x02&\xc0\x06<\x18\x0267\x18\x0234\x16\x00(\x0267\x18\x0234\x00\x00\x16\x94=\x156\x16\xd4.\x15F\x00&\xb4\t\x1c\x15\x0c\x195\x06\x08\x00\x19\x18\x0e__extracted_at\x15\x0c\x16\x06\x16\xd8\x07\x16\xa6\x08&\xb4\t<\x18\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x16\x00(\x192024-12-04T12:05:25+01:00\x18\x192024-12-04T12:05:25+01:00\x00\x00\x16\xca=\x15<\x16\x9a/\x15\xda\x02\x00&\xda\x11\x1c\x15\x0c\x195\x06\x08\x00\x19\x18\n__filepath\x15\x0c\x16\x06\x16\x80\x1b\x16\xce\x1b&\xda\x11<\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x16\x00(k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x18k/drives/b!MQ9fxJoWKE6kuaOULF7bUw7DQKsh_IlJhBdZh6ZKbeRkHtCdmELYTaVH26RsQ1EZ/root:/Connector Test/Test 1.xlsx\x00\x00\x16\x86>\x15<\x16\xf41\x15\xb2\n\x00\x16\x9a*\x16\x06\x00\x00\t\x04\x00\x00PAR1"
        request = mock.Mock(
            method="GET",
            url=mock.Mock(scheme="http", path="/", query=b"foo=bar"),
            headers={
                "x-neos-date": "20230101T120000Z",
                "x-neos-content-sha256": "f4eb19f40510b16354e25f8b339dca7a40e44dfb846214371c054677c42146d7",
                "Authorization": "NEOS4-HMAC-SHA256 Credential=access-key/20230101/ksa/iam/neos4_request, SignedHeaders=x-neos-date, Signature=signature",
            },
            body=AsyncMock(return_value=body),
        )

        v = validator.SignatureValidator(mock_hub)

        await v.validate(request, [action], [resource], logic_operator="and")

        assert mock_hub.validate_signature.call_args == mock.call(
            "access-key",
            "NEOS4",
            "20230101/ksa/iam/neos4_request",
            "NEOS4-HMAC-SHA256\n20230101T120000Z\n20230101/ksa/iam/neos4_request\n6165f63a8c53a61e1e05933269db3eca53b2bbe6ef1d359057c5dd27dbe4d66b",
            "signature",
            [action],
            [resource.urn],
            logic_operator="and",
            return_allowed_resources=False,
        )



from unittest import mock

import pytest

cli = pytest.importorskip("neos_common.cli")


def test_routes_printer():
    fastapi_mock = mock.Mock(
        routes=[
            mock.Mock(
                path="/test",
                methods=["GET"],
                summary="Test endpoint",
                openapi_extra={
                    "x-iam-action": "do:test",
                    "x-iam-resource": "urn:test",
                },
            ),
            mock.Mock(
                path="/ignore",
                methods=["GET"],
                summary="Ignored endpoint",
                openapi_extra={
                    "x-iam-action": "do:test",
                    "x-iam-resource": "urn:test",
                },
            ),
        ],
    )

    echo_fn = mock.Mock()

    cli.RoutesPrinter(fastapi_mock, ["/ignore"]).echo(echo_fn)

    assert echo_fn.call_args_list == [
        mock.call("Test endpoint"),
        mock.call("GET       /test"),
        mock.call("            AND"),
        mock.call("              action:   do:test"),
        mock.call("              resource: urn:test"),
        mock.call(),
        mock.call(),
    ]


def test_routes_printer_no_permissions():
    fastapi_mock = mock.Mock(
        routes=[
            mock.Mock(
                path="/test",
                methods=["GET"],
                summary="Test endpoint",
                openapi_extra={},
            ),
            mock.Mock(
                path="/ignore",
                methods=["GET"],
                summary="Ignored endpoint",
                openapi_extra={
                    "x-iam-action": "do:test",
                    "x-iam-resource": "urn:test",
                },
            ),
        ],
    )

    echo_fn = mock.Mock()

    cli.RoutesPrinter(fastapi_mock, ["/ignore"]).echo(echo_fn)

    assert echo_fn.call_args_list == [
        mock.call("Test endpoint"),
        mock.call("GET       /test"),
        mock.call(),
    ]


def test_csv_routes_printer():
    fastapi_mock = mock.Mock(
        routes=[
            mock.Mock(
                path="/test",
                methods=["GET"],
                summary="Test endpoint",
                openapi_extra={
                    "x-iam-action": "do:test",
                    "x-iam-resource": "urn:test",
                },
            ),
            mock.Mock(
                path="/ignore",
                methods=["GET"],
                summary="Ignored endpoint",
                openapi_extra={
                    "x-iam-action": "do:test",
                    "x-iam-resource": "urn:test",
                },
            ),
        ],
    )

    echo_fn = mock.Mock()

    cli.RoutesCSVPrinter(fastapi_mock, ["/ignore"]).echo(echo_fn)

    assert echo_fn.call_args_list == [
        mock.call("methods,path,summary,operator,action,resource\r\nGET,/test,Test endpoint,AND,do:test,urn:test\r\n"),
    ]


def test_csv_routes_printer_no_permissions():
    fastapi_mock = mock.Mock(
        routes=[
            mock.Mock(
                path="/test",
                methods=["GET"],
                summary="Test endpoint",
                openapi_extra={},
            ),
            mock.Mock(
                path="/ignore",
                methods=["GET"],
                summary="Ignored endpoint",
                openapi_extra={
                    "x-iam-action": "do:test",
                    "x-iam-resource": "urn:test",
                },
            ),
        ],
    )

    echo_fn = mock.Mock()

    cli.RoutesCSVPrinter(fastapi_mock, ["/ignore"]).echo(echo_fn)

    assert echo_fn.call_args_list == [
        mock.call("methods,path,summary,operator,action,resource\r\nGET,/test,Test endpoint,,,\r\n"),
    ]



import uuid
from pathlib import Path
from unittest import mock

import pytest
from _pytest.capture import CaptureFixture

fastapi = pytest.importorskip("fastapi")
requests = pytest.importorskip("starlette.requests")
staticfiles = pytest.importorskip("starlette.staticfiles")
testclient = pytest.importorskip("starlette.testclient")
timing = pytest.importorskip("neos_common.middleware.timing")


@pytest.fixture
def client():
    app = fastapi.FastAPI()
    timing.add_timing_middleware(app, exclude="untimed")
    static_files_app = staticfiles.StaticFiles(directory=".")
    app.mount(path="/static", app=static_files_app, name="static")

    @app.get("/timed")
    def get_timed() -> None:
        pass

    @app.get("/untimed")
    def get_untimed() -> None:
        pass

    return testclient.TestClient(app)


def test_timing(capsys: CaptureFixture[str], client, monkeypatch) -> None:
    trace_id = uuid.uuid4()
    monkeypatch.setattr(timing.uuid, "uuid4", mock.Mock(return_value=trace_id))
    client.get("/timed")
    out, err = capsys.readouterr()
    assert err == ""
    assert out.startswith(f"TIMING ({trace_id.hex}): Start")
    assert f"TIMING ({trace_id.hex}): Wall" in out
    assert "CPU:" in out
    assert out.endswith("test_timing.get_timed\n")


def test_silent_timing(capsys: CaptureFixture[str], client) -> None:
    client.get("/untimed")
    out, err = capsys.readouterr()
    assert err == ""
    assert out == ""


def test_mount(capsys: CaptureFixture[str], client, monkeypatch) -> None:
    trace_id = uuid.uuid4()
    monkeypatch.setattr(timing.uuid, "uuid4", mock.Mock(return_value=trace_id))
    basename = Path(__file__).name
    client.get(f"/static/{basename}")
    out, err = capsys.readouterr()
    assert err == ""
    assert out.startswith(f"TIMING ({trace_id.hex}):")
    assert out.endswith("StaticFiles<'static'>\n")


def test_missing(capsys: CaptureFixture[str], client, monkeypatch) -> None:
    trace_id = uuid.uuid4()
    monkeypatch.setattr(timing.uuid, "uuid4", mock.Mock(return_value=trace_id))
    client.get("/will-404")
    out, err = capsys.readouterr()
    assert err == ""
    assert out.startswith(f"TIMING ({trace_id.hex}):")
    assert out.endswith("<Path: /will-404>\n")


@pytest.fixture
def client2():
    app2 = fastapi.FastAPI()
    timing.add_timing_middleware(app2, prefix="app2")

    @app2.get("/")
    def get_with_intermediate_timing(request: requests.Request) -> None:
        timing.record_timing(request, note="hello")

    return testclient.TestClient(app2)


def test_intermediate(capsys: CaptureFixture[str], client2, monkeypatch) -> None:
    trace_id = uuid.uuid4()
    monkeypatch.setattr(timing.uuid, "uuid4", mock.Mock(return_value=trace_id))
    client2.get("/")
    out, err = capsys.readouterr()
    assert err == ""
    outs = out.strip().split("\n")
    assert len(outs) == 3
    assert outs[0].startswith(f"TIMING ({trace_id.hex}): Start")
    assert outs[1].startswith(f"TIMING ({trace_id.hex}):")
    assert outs[1].endswith("test_timing.get_with_intermediate_timing (hello)")
    assert outs[2].startswith(f"TIMING ({trace_id.hex}):")
    assert outs[2].endswith("test_timing.get_with_intermediate_timing")


@pytest.fixture
def client3():
    app3 = fastapi.FastAPI()

    @app3.get("/")
    def fail_to_record(request: requests.Request) -> None:
        timing.record_timing(request)

    return testclient.TestClient(app3)


def test_recording_fails_without_middleware(client3) -> None:
    with pytest.raises(
        RuntimeError,
        match="No timer, or invalid timer, present on request. Ensure timing middleware is added to the app.",
    ):
        client3.get("/")
